#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gnome.h>

#include <strings.h>
#include <ctype.h>
#include <assert.h>
#include <math.h>

// lcms.h needed for:
//    INTENT_* symbols to match to intent combobox entries
//    tag signatures for filling in edit fields
//#include <lcms/lcms.h>

#include "piccel.h"
#include "piccel_defs.h"

#include <icc34.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "update_flags.h"

extern UL update_flags1,update_flags2;
extern double round(double);

extern GtkWidget *piccel_window1;

int SB_NO_Update=0;    // don't change the original curve just because we estimate the gamma

LPLUT workLUT=0;           // results of last clut (re)build operation
icTagSignature workTag=0;  // and tag for which clut we rebuilt

GtkButton *main_win_edit_btn=0;


#define PICCEL_MAX_CHANNELS 4

#define d(x) ((double) (x))

/*****************
* GLOBALS
*****************/

GtkWindow *piccelEditIPWindow = 0;
GtkWindow *piccelEditOPWindow = 0;
GtkWindow *piccelEditMPWindow = 0;

GtkWindow *piccelGenCLUTWindow = 0;

GdkColor gamutAlarmColor;


double satArray[2][6];
double lumArray[2][6];
double shiftArray[2][6];

double a_satArray[2][6];
double a_lumArray[2][6];
double a_shiftArray[2][6];


char errstr[256];

#define DIRLEN 512
char imagefile[DIRLEN];
char iprof[DIRLEN];
char oprof[DIRLEN];
char mprof[DIRLEN];
char *profilename=0;
char currdir[DIRLEN];
char lastinitfile[DIRLEN];

/**** SUPPORT
*****************/

static void
alert() {

  GtkWidget *adlg;
  adlg = create_alert();
  gtk_label_set_text((GtkLabel*) lookup_widget(adlg,"alertText"),errstr);
  gtk_dialog_run (GTK_DIALOG (adlg));
  gtk_widget_destroy(adlg);
}

void
setdir(GtkWidget *fdlg) {
  if (currdir[0]) {
    gtk_file_chooser_set_current_folder((GtkFileChooser *) fdlg,currdir);
  }
}

void 
savedir(char *filename) {
  char *ndx;
  strncpy(currdir,filename,DIRLEN);
  ndx = strrchr(currdir,'/');
  if (ndx) {
    *(++ndx)='\0';
  }
    
}

void
loadimage(char *imagefile,GtkWidget *widget) {
  GtkImage *image1, *image2;
  GdkPixbuf *img=0;

  printf ("loading image file %s\n", imagefile);
  if ((img = pLoadBaseImage (imagefile)) != (GdkPixbuf *) NULL) {
    //fprintf(stderr,"file opened\n");
    image1 = (GtkImage *) lookup_widget (widget,"image1");
    //fprintf(stderr,"image1 found\n");
    gtk_image_set_from_pixbuf (image1, img);
    //fprintf(stderr,"img set 1\n");
    image2 = (GtkImage *) lookup_widget (widget,"image2");
    //fprintf(stderr,"image2 found\n");
    img = pLoadBaseImage (imagefile);
    gtk_image_set_from_pixbuf (image2, img);
    //fprintf(stderr,"img set 2\n");
    //g_free (img);
    //fprintf(stderr,"free and finished\n");

  }
}


/** lcms - combobox
*/
int
convertIntentCB (int i)
{
	switch (i)
	{
	case 0:
		return INTENT_PERCEPTUAL;
	case 1:
		return INTENT_RELATIVE_COLORIMETRIC;
	case 2:
		return INTENT_SATURATION;
	case 3:
		return INTENT_ABSOLUTE_COLORIMETRIC;
	default:
		fprintf (stderr, "convertIntentCB: %d not identified\n", i);
		return INTENT_PERCEPTUAL;
	}
}

int
convertIntentLCMS (int i)
{
	switch (i)
	{
	case INTENT_PERCEPTUAL:
		return 0;
	case INTENT_RELATIVE_COLORIMETRIC:
		return 1;
	case INTENT_SATURATION:
		return 2;
	case INTENT_ABSOLUTE_COLORIMETRIC:
		return 3;
	default:
		fprintf (stderr, "convertIntentLCMS: %d not identified\n", i);
		return 0;
	}
}

int
convertDeviceClassSig (icProfileClassSignature sig)
{
	switch (sig)
	{
	case icSigInputClass:
		return 0;
	case icSigDisplayClass:
		return 1;
	case icSigOutputClass:
		return 2;
	case icSigLinkClass:
		return 3;
	case icSigAbstractClass:
		return 4;
	case icSigColorSpaceClass:
		return 5;
	case icSigNamedColorClass:
		return 6;
	default:
		fprintf (stderr,
			 "convertDeviceClassSig: icProfileClassSignature %0x not recognised\n",
			 sig);
		return 0;
	}
}

icProfileClassSignature
convertDeviceClassCB (int ndx)
{
	switch (ndx)
	{
	case 0:
		return icSigInputClass;
	case 1:
		return icSigDisplayClass;
	case 2:
		return icSigOutputClass;
	case 3:
		return icSigLinkClass;
	case 4:
		return icSigAbstractClass;
	case 5:
		return icSigColorSpaceClass;
	case 6:
		return icSigNamedColorClass;
	default:
		fprintf (stderr,
			 "convertDeviceClassCB: combobox entry %d not recognised\n",
			 ndx);
		return icSigInputClass;
	}
}

int
convertColSpaceSig (icColorSpaceSignature sig)
{
	switch (sig)
	{
	case icSigXYZData:
		return 0;
	case icSigLabData:
		return 1;
	case icSigLuvData:
		return 2;
	case icSigYCbCrData:
		return 3;
	case icSigYxyData:
		return 4;
	case icSigRgbData:
		return 5;
	case icSigGrayData:
		return 6;
	case icSigHsvData:
		return 7;
	case icSigHlsData:
		return 8;
	case icSigCmykData:
		return 9;
	case icSigCmyData:
		return 10;
	case icSig2colorData:
		return 11;
	case icSig3colorData:
		return 12;
	case icSig4colorData:
		return 13;
	case icSig5colorData:
		return 14;
	case icSig6colorData:
		return 15;
	case icSig7colorData:
		return 16;
	case icSig8colorData:
		return 17;
	case icSig9colorData:
		return 18;
	case icSig10colorData:
		return 19;
	case icSig11colorData:
		return 20;
	case icSig12colorData:
		return 21;
	case icSig13colorData:
		return 22;
	case icSig14colorData:
		return 23;
	case icSig15colorData:
		return 24;
	default:
		fprintf (stderr,
			 "convertColSpaceSig: icColorSpaceSignature %0x not recognised\n",
			 sig);
		return 0;
	}

}

icColorSpaceSignature
convertColSpaceCB (int ndx)
{
	switch (ndx)
	{
	case 0:
		return icSigXYZData;
	case 1:
		return icSigLabData;
	case 2:
		return icSigLuvData;
	case 3:
		return icSigYCbCrData;
	case 4:
		return icSigYxyData;
	case 5:
		return icSigRgbData;
	case 6:
		return icSigGrayData;
	case 7:
		return icSigHsvData;
	case 8:
		return icSigHlsData;
	case 9:
		return icSigCmykData;
	case 10:
		return icSigCmyData;
	case 11:
		return icSig2colorData;
	case 12:
		return icSig3colorData;
	case 13:
		return icSig4colorData;
	case 14:
		return icSig5colorData;
	case 15:
		return icSig6colorData;
	case 16:
		return icSig7colorData;
	case 17:
		return icSig8colorData;
	case 18:
		return icSig9colorData;
	case 19:
		return icSig10colorData;
	case 20:
		return icSig11colorData;
	case 21:
		return icSig12colorData;
	case 22:
		return icSig13colorData;
	case 23:
		return icSig14colorData;
	case 24:
		return icSig15colorData;
	default:
		fprintf (stderr,
			 "convertColSpaceCB: combobox index %d not recognised\n",
			 ndx);
		return icSigXYZData;
	}

}






/** Gtk widget helpers
*/
void
loadEntryWithUInt (GtkWidget * widget, char *eName, unsigned int d)
{
	char text[32] = "\0\0\0";
	GtkEntry *entry;

	entry = (GtkEntry *) lookup_widget (widget, eName);
	sprintf (text, "%u", d);
	gtk_entry_set_text (entry, (const gchar *) text);
}


void
hideWidget (GtkWidget * otherWidget, char *wName)
{
	GtkWidget *widget;
	widget = lookup_widget (otherWidget, wName);
	gtk_widget_hide (widget);
}

int
readEntryDouble(GtkWidget * widget, char *eName, double *rslt) {
  char *str;

  str = (char *) gtk_entry_get_text((GtkEntry *) lookup_widget (widget, eName));
  if ((str == (char *)NULL) || (!strlen(str))) 
    return FALSE;

  sscanf(str,"%lf",rslt);
  return TRUE;
}

void
loadEntryWithDouble (GtkWidget * widget, char *eName, double d)
{
	char dspace[64] = "\0\0\0";
	GtkEntry *entry;

	entry = (GtkEntry *) lookup_widget (widget, eName);
	sprintf (dspace, "%lf", d);
	gtk_entry_set_text (entry, (const gchar *) dspace);
}

void
loadEWD3 (double darr[3], char *carr[3], GtkWidget * widget)
{
	int i;

	for (i = 0; i < 3; i++)
	{
		loadEntryWithDouble (widget, carr[i], darr[i]);
	}

}



/** window init routintes
*/
void
initGamutAlarmColor ()
{
	gamutAlarmColor.red = 0x00ff;
	gamutAlarmColor.green = 0x00ff;
	gamutAlarmColor.blue = 0x00ff;

	cmsSetAlarmCodes (gamutAlarmColor.red, gamutAlarmColor.green,
			  gamutAlarmColor.blue);
}

void
initIntent (char pid, GtkWidget * widget)
{
	GtkComboBox *cbx;
	char cbname[16];

	switch (pid)
	{
	case 'i':
		strcpy (cbname, "ip_intent");
		break;
	case 'o':
		strcpy (cbname, "op_intent");
		break;
	case 'm':
		strcpy (cbname, "mp_intent");
		break;
	default:
		fprintf (stderr, "initIntent: bad profile id code %c\n", pid);
		return;
		break;
	}

	cbx = (GtkComboBox *) lookup_widget (widget, cbname);
	gtk_combo_box_set_active (cbx, 0);

}


/**** PICCEL MAIN WINDOW
*
*****************/
/*** MENUs 
*
**/
/* File menu
*/

int 
loadstate(char *filename) {
  FILE *initfile;
  fprintf(stderr,"loadstate %s\n",filename);

  if ((initfile = fopen(filename,"r")) != (FILE *) NULL) {
    char *inlin=NULL;
    int len=0;

    while (getline(&inlin,&len,initfile) >=0 ) {
      //fprintf(stderr,"len %d line is: .%s.\n",len,inlin);
      if ((sscanf(inlin,"IMAGE:%s \n",imagefile)) >0) {
	loadimage(imagefile,piccel_window1);
      } else if ((sscanf(inlin,"INPROF:%s \n",iprof))>0) {
	printf("loading input profile %s\n",iprof);
	profilename = pLoadInputProfile(iprof);
	if (profilename) {
	  printf ("profile is .%s.\n", profilename);
	  setProfileName ('i', profilename, GTK_WIDGET (piccel_window1));
	} else {
	  setProfileName ('i', iprof, GTK_WIDGET (piccel_window1));
	}
      } else if ((sscanf(inlin,"OUTPROF:%s",oprof))>0) {
	printf("loading output profile %s\n",oprof);
	profilename = pLoadOutputProfile(oprof);
	if (profilename) {
	  printf ("profile is .%s.\n", profilename);
	  setProfileName ('o', profilename, GTK_WIDGET (piccel_window1));
	} else {
	  setProfileName ('o', iprof, GTK_WIDGET (piccel_window1));
	}
      } else if ((fscanf(initfile,"MONPROF:%s \n",mprof))>0) {
	printf("loading monitor profile %s\n",iprof);
	profilename = pLoadMonitorProfile(mprof);
	if (profilename) {
	  printf ("profile is .%s.\n", profilename);
	  setProfileName ('m', profilename, GTK_WIDGET (piccel_window1));
	} else {
	  setProfileName ('m', iprof, GTK_WIDGET (piccel_window1));
	}
      } else if ((sscanf(inlin,"CURRDIR:%s \n",currdir))>0) {
      }
    }
    if (inlin) 
      free(inlin);

    
    fclose(initfile);
    strncpy(lastinitfile,filename,DIRLEN);
    return 1;
  } else {
    fprintf(stderr,"failed to open %s for read\n",filename);
    return 0;
  }
}

int
savestate(char *filename) {
  FILE *initfile;

  if ((initfile = fopen(filename,"w")) != (FILE *) NULL) {
    fprintf(initfile,"IMAGE:%s \n",imagefile);
    fprintf(initfile,"INPROF:%s \n",iprof);
    fprintf(initfile,"OUTPROF:%s \n",oprof);
    fprintf(initfile,"MONPROF:%s \n",mprof);
    fprintf(initfile,"CURRDIR:%s \n",currdir);
    
    fclose(initfile);
    strncpy(lastinitfile,filename,DIRLEN);
    return 1;
  } else {
    fprintf(stderr,"failed to open %s for write\n",filename);
    return 0;
  }

}

void
on_new1_activate (GtkMenuItem * menuitem, gpointer user_data)
{

}


void
on_open1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  GtkWidget *fcdlg;
  char *filename=0;
  fcdlg = create_filechooserdialog1 ();
  setdir(fcdlg);
  if (gtk_dialog_run (GTK_DIALOG (fcdlg)) == GTK_RESPONSE_OK)
    {
      filename =
	gtk_file_chooser_get_filename (GTK_FILE_CHOOSER
				       (fcdlg));
      savedir(filename);
      loadstate(filename);
    }
  else
    {
      printf ("load init file: response not OK\n");
    }
  
  gtk_widget_destroy (fcdlg);
  
}


void
on_save1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  savestate(lastinitfile);
}


void
on_save_as1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  GtkWidget *fsdlg;
  char *filename=0;
  fsdlg = create_filesaverdialog();
  setdir(fsdlg);
  if (gtk_dialog_run (GTK_DIALOG (fsdlg)) == GTK_RESPONSE_OK)
    {
      filename =
	gtk_file_chooser_get_filename (GTK_FILE_CHOOSER
				       (fsdlg));
      //savedir(filename);
      savestate(filename);
    }
  else
    {
      printf ("save as: response not OK\n");
    }
  
  gtk_widget_destroy (fsdlg);
  
}


void
on_quit1_activate (GtkMenuItem * menuitem, gpointer user_data)
{
  gtk_widget_destroy(lookup_widget((GtkWidget*) menuitem,"piccel_window1"));
}

/* Edit menu
*/

void
on_cut1_activate (GtkMenuItem * menuitem, gpointer user_data)
{

}


void
on_copy1_activate (GtkMenuItem * menuitem, gpointer user_data)
{

}


void
on_paste1_activate (GtkMenuItem * menuitem, gpointer user_data)
{

}


void
on_clear1_activate (GtkMenuItem * menuitem, gpointer user_data)
{

}


void
on_properties1_activate (GtkMenuItem * menuitem, gpointer user_data)
{

}


void
on_preferences1_activate (GtkMenuItem * menuitem, gpointer user_data)
{

}

/* Help menu
*/

void
on_about1_activate (GtkMenuItem * menuitem, gpointer user_data)
{

}


/*** Image Frame 
*/


void
on_piccel_window1_destroy (GtkObject * object, gpointer user_data)
{
	char *title;

	title = (char *) gtk_window_get_title (GTK_WINDOW (object));
	fprintf (stderr, "closing edit window %s\n", title);

	pCloseProfiles ();
	exit (0);

}

gboolean
on_image1_expose_event (GtkWidget * widget,
			GdkEventExpose * event, gpointer user_data)
{
	// supposed to scroll image2 when image1 is scrolled, but 
	// why does it scroll image1 when image2 is scrolled?

	//printf("image1 expose\n");

	GtkScrolledWindow *sw2, *sw1;

	sw1 = (GtkScrolledWindow *) lookup_widget (GTK_WIDGET (widget),
						   "scrolledwindow1");
	sw2 = (GtkScrolledWindow *) lookup_widget (GTK_WIDGET (widget),
						   "scrolledwindow2");

	gtk_scrolled_window_set_hadjustment (sw2,
					     gtk_scrolled_window_get_hadjustment
					     (sw1));
	gtk_scrolled_window_set_vadjustment (sw2,
					     gtk_scrolled_window_get_vadjustment
					     (sw1));

	return FALSE;
}

void
on_load_button_clicked (GtkButton * button, gpointer user_data)
{
	GtkWidget *fcdlg;
	char *filename;

	fcdlg = create_filechooserdialog1 ();
	setdir(fcdlg);
	if (gtk_dialog_run (GTK_DIALOG (fcdlg)) == GTK_RESPONSE_OK)
	{
		filename =
			gtk_file_chooser_get_filename (GTK_FILE_CHOOSER
						       (fcdlg));
		strncpy(imagefile,filename,DIRLEN);
		savedir(imagefile);
		loadimage(imagefile,GTK_WIDGET(button));
	}
	else
	{
		printf ("load image: response not OK\n");
	}

	gtk_widget_destroy (fcdlg);



}


void
on_image_info_clicked (GtkButton * button, gpointer user_data)
{
	GtkTextBuffer *text_buffer;
	GtkTextView *i_info_text;
	GtkWidget *image_info, *image1;
	char iit[2048];
	GdkPixbuf *i1pixbuf;
	GtkTextIter iter;

	image1 = lookup_widget (GTK_WIDGET (button), "image1");
	i1pixbuf = gtk_image_get_pixbuf (GTK_IMAGE (image1));
	if (i1pixbuf != (GdkPixbuf *) NULL)
	{
		sprintf (iit,
			 " height:               %6d \n width:                 %6d \n channels:                 %2d \n colorspace:      %8s \n bits per sample:       %2d \n has alpha:           %6s \n rowstride:        %8d\n",
			 gdk_pixbuf_get_height (i1pixbuf),
			 gdk_pixbuf_get_width (i1pixbuf),
			 gdk_pixbuf_get_n_channels (i1pixbuf),
			 (gdk_pixbuf_get_colorspace (i1pixbuf) ==
			  GDK_COLORSPACE_RGB ? "RGB" : "not RGB"),
			 gdk_pixbuf_get_bits_per_sample (i1pixbuf),
			 (gdk_pixbuf_get_has_alpha (i1pixbuf) ? "true" :
			  "false"), gdk_pixbuf_get_rowstride (i1pixbuf));
	}

	image_info = create_info_dialog ();

	i_info_text = (GtkTextView *) lookup_widget (image_info, "info_text");
	text_buffer = gtk_text_view_get_buffer (i_info_text);

	gtk_text_buffer_create_tag (text_buffer, "bold",
				    "weight", PANGO_WEIGHT_BOLD, NULL);
	gtk_text_buffer_create_tag (text_buffer, "right_justify",
				    "justification", GTK_JUSTIFY_RIGHT, NULL);

	gtk_text_buffer_get_iter_at_offset (text_buffer, &iter, 0);

	gtk_text_buffer_insert_with_tags_by_name (text_buffer, &iter,
						  "height:", -1, "bold",
						  NULL);
	sprintf (iit, "                % 8d \n",
		 gdk_pixbuf_get_height (i1pixbuf));
	gtk_text_buffer_insert (text_buffer, &iter, iit, -1);

	gtk_text_buffer_insert_with_tags_by_name (text_buffer, &iter,
						  "width:", -1, "bold", NULL);
	sprintf (iit, "                 % 8d \n",
		 gdk_pixbuf_get_width (i1pixbuf));
	gtk_text_buffer_insert_with_tags_by_name (text_buffer, &iter, iit, -1,
						  "right_justify", NULL);

	gtk_text_buffer_insert_with_tags_by_name (text_buffer, &iter,
						  "channels:", -1, "bold",
						  NULL);
	sprintf (iit, "              % 8d \n",
		 gdk_pixbuf_get_n_channels (i1pixbuf));
	gtk_text_buffer_insert (text_buffer, &iter, iit, -1);

	gtk_text_buffer_insert_with_tags_by_name (text_buffer, &iter,
						  "colorspace:", -1, "bold",
						  NULL);
	sprintf (iit, "       %8s \n",
		 (gdk_pixbuf_get_colorspace (i1pixbuf) ==
		  GDK_COLORSPACE_RGB ? "RGB" : "not RGB"));
	gtk_text_buffer_insert (text_buffer, &iter, iit, -1);

	gtk_text_buffer_insert_with_tags_by_name (text_buffer, &iter,
						  "bits per sample", -1,
						  "bold", NULL);
	sprintf (iit, "  % 8d \n", gdk_pixbuf_get_bits_per_sample (i1pixbuf));
	gtk_text_buffer_insert (text_buffer, &iter, iit, -1);

	gtk_text_buffer_insert_with_tags_by_name (text_buffer, &iter,
						  "has alpha:", -1, "bold",
						  NULL);
	sprintf (iit, "            %6s \n",
		 (gdk_pixbuf_get_has_alpha (i1pixbuf) ? "true" : "false"));
	gtk_text_buffer_insert (text_buffer, &iter, iit, -1);

	gtk_text_buffer_insert_with_tags_by_name (text_buffer, &iter,
						  "rowstride:", -1, "bold",
						  NULL);
	sprintf (iit, "          % 8d\n",
		 gdk_pixbuf_get_rowstride (i1pixbuf));
	gtk_text_buffer_insert (text_buffer, &iter, iit, -1);

	gtk_dialog_run (GTK_DIALOG (image_info));

	gtk_widget_destroy (image_info);

}


void
on_image_update_clicked (GtkButton * button, gpointer user_data)
{
	GtkImage *image1, *image2;
	GdkPixbuf *i1pb, *i2pb;
	GtkToggleButton *ckbtn;
	GtkComboBox *cbox;

	int ip, op, mp, gamut, intent, proofIntent;
	int ipIntent, opIntent, mpIntent;

	ckbtn = (GtkToggleButton *) lookup_widget (GTK_WIDGET (button),
						   "ip_apply");
	ip = gtk_toggle_button_get_active (ckbtn);
	ckbtn = (GtkToggleButton *) lookup_widget (GTK_WIDGET (button),
						   "op_apply");
	op = gtk_toggle_button_get_active (ckbtn);
	ckbtn = (GtkToggleButton *) lookup_widget (GTK_WIDGET (button),
						   "mp_apply");
	mp = gtk_toggle_button_get_active (ckbtn);

	ckbtn = (GtkToggleButton *) lookup_widget (GTK_WIDGET (button),
						   "op_gamut_chk");
	gamut = gtk_toggle_button_get_active (ckbtn);

	cbox = (GtkComboBox *) lookup_widget (GTK_WIDGET (button),
					      "ip_intent");
	ipIntent = convertIntentCB (gtk_combo_box_get_active (cbox));
	cbox = (GtkComboBox *) lookup_widget (GTK_WIDGET (button),
					      "op_intent");
	opIntent = convertIntentCB (gtk_combo_box_get_active (cbox));
	proofIntent = opIntent;
	cbox = (GtkComboBox *) lookup_widget (GTK_WIDGET (button),
					      "mp_intent");
	mpIntent = convertIntentCB (gtk_combo_box_get_active (cbox));


	if (op && mp)
	{
		intent = mpIntent;
		proofIntent = opIntent;
	}
	else if (mp)
	{
		intent = mpIntent;
	}
	else if (op)
	{
		intent = opIntent;
	}
	else if (ip)
	{
		intent = ipIntent;
	}
	else
	{
		intent = INTENT_PERCEPTUAL;
	}

	fprintf (stderr, "building transform\n");
	pBuildTransform (ip, op, mp, gamut, intent, proofIntent);

	image1 = (GtkImage *) lookup_widget (GTK_WIDGET (button), "image1");
	image2 = (GtkImage *) lookup_widget (GTK_WIDGET (button), "image2");

	i1pb = gtk_image_get_pixbuf (image1);
	i2pb = gtk_image_get_pixbuf (image2);

	fprintf (stderr, "applying transform\n");
	pApplyTransform (i1pb, i2pb);

	fprintf (stderr, "rendering image\n");
	gtk_widget_queue_draw (GTK_WIDGET (image2));
	fprintf (stderr, "done.\n");
}



/*** Profile frames
*/
/** support routines
*/
void
setProfileName (char pid, char *profileName, GtkWidget * widget)
{
	char twname[16];
	GtkEntry *textEntry;

	switch (pid)
	{
	case 'i':
		strcpy (twname, "ip_text");
		break;
	case 'o':
		strcpy (twname, "op_text");
		break;
	case 'm':
		strcpy (twname, "mp_text");
		break;
	default:
		fprintf (stderr, "setProfileName: bad profile id code .%c.\n",
			 pid);
		return;
		break;
	}

	textEntry = (GtkEntry *) lookup_widget (GTK_WIDGET (widget), twname);
	gtk_entry_set_text (textEntry, profileName);

}



void
set_profile (char pid, GtkButton * button)
{
	GtkWidget *fcdlg;

	fcdlg = create_filechooserdialog1 ();
	setdir(fcdlg);
	if (gtk_dialog_run (GTK_DIALOG (fcdlg)) == GTK_RESPONSE_OK)
	{
		char *filename;
		char *fnp;

		filename =
			gtk_file_chooser_get_filename (GTK_FILE_CHOOSER
						       (fcdlg));
		savedir(filename);

		switch (pid)
		{
		case 'i':
			printf ("loading input profile: %s\n", filename);
			strncpy(iprof,filename,DIRLEN);
			profilename = pLoadInputProfile (filename);
			break;
		case 'o':
			printf ("loading output profile: %s\n", filename);
			strncpy(oprof,filename,DIRLEN);
			profilename = pLoadOutputProfile (filename);
			break;
		case 'm':
			printf ("loading monitor profile: %s\n", filename);
			strncpy(mprof,filename,DIRLEN);
			profilename = pLoadMonitorProfile (filename);
			break;
		default:
			fprintf (stderr, "set_profile: bad profile code %c\n",
				 pid);
			return;
			break;
		}

		if (profilename)
		{
			printf ("profile is .%s.\n", profilename);
			setProfileName (pid, profilename,
					GTK_WIDGET (button));
		}
		else
		{
			if ((fnp = strrchr (filename, '/')))
			{
				fnp++;
				setProfileName (pid, fnp,
						GTK_WIDGET (button));
			}
			else
			{
				setProfileName (pid, filename,
						GTK_WIDGET (button));
			}
		}

		g_free (filename);
	}
	else
	{
		printf ("set profile: response not OK\n");
	}

	gtk_widget_destroy (fcdlg);

}

void
show_profile_info (char pid, GtkWidget * widget)
{

	GtkTextBuffer *text_buffer;
	GtkTextView *info_text;
	GtkWidget *profile_info;
	GtkTextIter iter;


	//genICCinfo(pid,pit);

	profile_info = create_info_dialog ();

	info_text = (GtkTextView *) lookup_widget (profile_info, "info_text");
	text_buffer = gtk_text_view_get_buffer (info_text);

	gtk_text_buffer_create_tag (text_buffer, "bold",
				    "weight", PANGO_WEIGHT_BOLD, NULL);

	gtk_text_buffer_get_iter_at_offset (text_buffer, &iter, 0);

	gtk_text_buffer_insert_with_tags_by_name (text_buffer, &iter,
						  "LCMS product name: ", -1,
						  "bold", NULL);
	gtk_text_buffer_insert (text_buffer, &iter, getICC_ProdName (pid),
				-1);
	gtk_text_buffer_insert_with_tags_by_name (text_buffer, &iter,
						  "\nLCMS profile info: \n\n",
						  -1, "bold", NULL);
	gtk_text_buffer_insert (text_buffer, &iter, getICC_ProfInfo (pid),
				-1);

	gtk_dialog_run (GTK_DIALOG (profile_info));
	gtk_widget_destroy (profile_info);


}


/** Profile frames -- callbacks
*/
/* load profile
*/
void
on_load_ip_clicked (GtkButton * button, gpointer user_data)
{
	set_profile ('i', button);
}

void
on_load_op_clicked (GtkButton * button, gpointer user_data)
{
	set_profile ('o', button);
}

void
on_load_mp_clicked (GtkButton * button, gpointer user_data)
{
	set_profile ('m', button);
}




/* edit profile
*/
/* prototypes for edit window */ 
void loadEditWinInfoTab (char pid);
void loadEditWinAtoBTab (char pid);
void loadEditWinGamutTab (char pid);
void loadEditWinTRCTab (char pid);
void loadEditWinPreviewTab (char pid);

void
on_edit_ip_clicked (GtkButton * button, gpointer user_data)
{
	if (piccelEditIPWindow != 0)
		return;

	main_win_edit_btn = button;

	piccelEditIPWindow = (GtkWindow *) create_piccel_edit_window2 ();
	gtk_window_set_title (piccelEditIPWindow,
			      (const gchar *) "piccel edit input profile");

	loadEditWinInfoTab ('i');
	loadEditWinAtoBTab ('i');
	loadEditWinGamutTab ('i');
	loadEditWinTRCTab ('i');
	loadEditWinPreviewTab ('i');
	
	gtk_widget_show (GTK_WIDGET (piccelEditIPWindow));


}


void
on_edit_op_clicked (GtkButton * button, gpointer user_data)
{
	if (piccelEditOPWindow == 0)
	  piccelEditOPWindow = (GtkWindow *) create_piccel_edit_window2 ();

	main_win_edit_btn = button;

	gtk_window_set_title (piccelEditOPWindow,
			      (const gchar *) "piccel edit output profile");
	loadEditWinInfoTab ('o');
	loadEditWinAtoBTab ('o');
	loadEditWinGamutTab ('o');
	loadEditWinTRCTab ('o');
	loadEditWinPreviewTab ('o');

	gtk_widget_show (GTK_WIDGET (piccelEditOPWindow));

	//setCurveTypes('o',"a2b0L1c1Curve");
}


void
on_edit_mp_clicked (GtkButton * button, gpointer user_data)
{
	if (piccelEditMPWindow != 0)
		return;

	main_win_edit_btn = button;

	piccelEditMPWindow = (GtkWindow *) create_piccel_edit_window2 ();
	gtk_window_set_title (piccelEditMPWindow,
			      (const gchar *) "piccel edit monitor profile");
	loadEditWinInfoTab ('m');
	loadEditWinAtoBTab ('m');
	loadEditWinGamutTab ('m');
	loadEditWinTRCTab ('m');
	loadEditWinPreviewTab ('m');

	gtk_widget_show (GTK_WIDGET (piccelEditMPWindow));
}

/* set sRGB profile
*/
void
on_ip_srgb_clicked (GtkButton * button, gpointer user_data)
{
	setProfileName ('i', pProfilesRGB ('i'), GTK_WIDGET (button));
}


void
on_op_srgb_clicked (GtkButton * button, gpointer user_data)
{
	setProfileName ('o', pProfilesRGB ('o'), GTK_WIDGET (button));
}


void
on_mp_srgb_clicked (GtkButton * button, gpointer user_data)
{
	setProfileName ('m', pProfilesRGB ('m'), GTK_WIDGET (button));
}

/* profile info
*/
void
on_ip_info_clicked (GtkButton * button, gpointer user_data)
{
	prtICCinfo ('i');
	show_profile_info ('i', GTK_WIDGET (button));
}


void
on_op_info_clicked (GtkButton * button, gpointer user_data)
{
	prtICCinfo ('o');
	show_profile_info ('o', GTK_WIDGET (button));
}


void
on_mp_info_clicked (GtkButton * button, gpointer user_data)
{
	prtICCinfo ('m');
	show_profile_info ('m', GTK_WIDGET (button));
}

/* output profile gamut color select
*/
void
on_op_gamut_color_clicked (GtkButton * button, gpointer user_data)
{
	GtkWidget *csdlg;
	GtkColorSelection *csel;

	csdlg = create_colorselectiondialog1 ();
	csel = (GtkColorSelection *) ((GtkColorSelectionDialog *) csdlg)->
		colorsel;
	gtk_color_selection_set_current_color (csel, &gamutAlarmColor);

	printf ("gamut alarm color start = r: %d  g: %d  b: %d \n",
			gamutAlarmColor.red, gamutAlarmColor.green, gamutAlarmColor.blue);

	
	if (gtk_dialog_run (GTK_DIALOG (csdlg)) == GTK_RESPONSE_OK)
	{
		gtk_color_selection_get_current_color (csel, &gamutAlarmColor);
		printf ("gamut alarm color = r: %d  g: %d  b: %d \n",
			gamutAlarmColor.red, gamutAlarmColor.green, gamutAlarmColor.blue);
		cmsSetAlarmCodes (gamutAlarmColor.red, gamutAlarmColor.green,
				  gamutAlarmColor.blue);
	}

	gtk_widget_destroy (csdlg);
}





/**** PICCEL EDIT WINDOW
*
*****************/
/** support routines
*/
/* Edit info tab support 
*/
void
loadCharTargetData (char pid, GtkWidget * window)
{
	GtkTextView *tview;
	GtkTextBuffer *tbuf;
	char *text;
	size_t len;

	if (!getCharTargetData (pid, &text, &len))
		return;

	tview = (GtkTextView *) lookup_widget (window, "charTargTextView");
	tbuf = gtk_text_view_get_buffer (tview);
	gtk_text_buffer_set_text (tbuf, text, (int) len);
}

void
loadColorants (char pid, GtkWidget * widget)
{
	double Rcols[3], Gcols[3], Bcols[3];
	char *eNames[3];

	if (!getColorants (pid, Rcols, Gcols, Bcols))
		return;

	eNames[0] = "redColXEntry";
	eNames[1] = "redColYEntry";
	eNames[2] = "redColZEntry";
	loadEWD3 (Rcols, eNames, widget);

	eNames[0] = "greenColXEntry";
	eNames[1] = "greenColYEntry";
	eNames[2] = "greenColZEntry";
	loadEWD3 (Gcols, eNames, widget);

	eNames[0] = "blueColXEntry";
	eNames[1] = "blueColYEntry";
	eNames[2] = "blueColZEntry";
	loadEWD3 (Bcols, eNames, widget);

}

void
loadMPt (char targ, char pid, GtkWidget * widget)
{
	double cieXYZ[3];
	char *eNames[3];

	if (!getMediaPt (targ, pid, cieXYZ))
		return;

	switch (targ)
	{
	case 'w':
		eNames[0] = "mWPXEntry";
		eNames[1] = "mWPYEntry";
		eNames[2] = "mWPZEntry";
		break;
	case 'b':
		eNames[0] = "mBPXEntry";
		eNames[1] = "mBPYEntry";
		eNames[2] = "mBPZEntry";
		break;
	default:
		fprintf (stderr, "loadMPt: bad target ID %c\n", targ);
		return;
		break;
	}

	loadEWD3 (cieXYZ, eNames, widget);
}

void
loadDevClass (char pid, GtkWidget * window)
{
	GtkComboBox *cb;
	cb = (GtkComboBox *) lookup_widget (window, "devClassCB");
	gtk_combo_box_set_active (cb,
				  convertDeviceClassSig (getDeviceClass
							 (pid)));
}

void
loadColorSpace (char pid, GtkWidget * window)
{
	GtkComboBox *cb;
	cb = (GtkComboBox *) lookup_widget (window, "colSpaceCB");
	gtk_combo_box_set_active (cb, convertColSpaceSig (getColSpace (pid)));
}

void
loadPCS (char pid, GtkWidget * window)
{
	GtkComboBox *cb;
	cb = (GtkComboBox *) lookup_widget (window, "pcsCB");
	gtk_combo_box_set_active (cb, convertColSpaceSig (getPCS (pid)));
}

void
loadIntent (char pid, GtkWidget * window)
{
	GtkComboBox *cb;
	cb = (GtkComboBox *) lookup_widget (window, "editIntentCB");
	gtk_combo_box_set_active (cb, convertIntentLCMS (getIntent (pid)));
}

void
loadProfileId (char pid, GtkWidget * window)
{
	char text[32] = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
	DWORD ver;
	LPBYTE profID;
	GtkEntry *entry;

	ver = getProfileICCversion (pid);
	if (ver >= 0x4000000)
	{
		int i;
		profID = getProfileICCid (pid);
		for (i = 0; i < 16; i++)
		{
			if (isprint (*profID))
				text[i] = *profID;
			profID++;
		}
		text[16] = '\0';
	}
	else
	{
		sprintf (text, "(not v4 profile)");
	}

	entry = (GtkEntry *) lookup_widget (window, "profileIDEntry");
	gtk_entry_set_text (entry, text);
}

GtkWidget *
pickEditWindow (char pid)
{
	switch (pid)
	{
	case 'i':
		return (GtkWidget *) piccelEditIPWindow;
		break;
	case 'o':
		return (GtkWidget *) piccelEditOPWindow;
		break;
	case 'm':
		return (GtkWidget *) piccelEditMPWindow;
		break;
	default:
		fprintf (stderr, "pickEditWndow: bad profile code %c\n", pid);
		return (GtkWidget *) 0;
		break;		// crash and burn
	}
}



/* Edit window info tab
*/
void
loadEditWinInfoTab (char pid)
{

	GtkEntry *entry;
	GtkWidget *window;

	if (!(window = pickEditWindow (pid)))
		return;

	loadCharTargetData (pid, window);

	entry = (GtkEntry *) lookup_widget (window, "copyrightEntry");
	gtk_entry_set_text (entry,
			    (const gchar *) getICCtagText (pid,
							   icSigCopyrightTag));

	entry = (GtkEntry *) lookup_widget (window, "profDescEntry");
	gtk_entry_set_text (entry,
			    (const gchar *) getICCtagText (pid,
							   icSigProfileDescriptionTag));

	entry = (GtkEntry *) lookup_widget (window, "devMfgEntry");
	gtk_entry_set_text (entry,
			    (const gchar *) getICCtagText (pid,
							   icSigDeviceMfgDescTag));

	entry = (GtkEntry *) lookup_widget (window, "devModelEntry");
	gtk_entry_set_text (entry,
			    (const gchar *) getICCtagText (pid,
							   icSigDeviceModelDescTag));


	loadColorants (pid, window);
	loadMPt ('w', pid, window);
	loadMPt ('b', pid, window);
	// loadChromaticity(pid,window);   // not implemented in lcms, though there is an 'addChromaticityTag'

	loadDevClass (pid, window);
	loadColorSpace (pid, window);
	loadPCS (pid, window);
	loadIntent (pid, window);
	loadProfileId (pid, window);
}

/* Edit A to B support routines
*/
void
loadCurveLx (GtkWidget * widget, char *cName, unsigned short *ldat,
	     unsigned int siz)
{
	gfloat *curveDat;
	unsigned int i;
	GtkCurve *curve;
	gfloat minX, maxX, minY, maxY;

	//fprintf(stderr,"loadcurveLx %s siz= %d \n",cName, siz);
	
	if (!(curveDat = malloc (siz * sizeof (float))))
	{
		fprintf (stderr, "unable to allocate memory for L curve\n");
		return;
	}

	minX = (gfloat) 0;
	maxX = (gfloat) siz;
	minY = (gfloat) 0;  //999999999.0;
	maxY = (gfloat) 65535; //- 999999999.0;

		//fprintf(stderr,"%s\n",cName);
	
	for (i = 0; i < siz; i++)
	{
		curveDat[i] = (gfloat) ldat[i];
		// fprintf(stderr,"i= %d cv= %d = %f   ldat= %d\n",i,(int) curveDat[i],curveDat[i],ldat[i]);

		/*
		if (curveDat[i] < minY)
			minY = curveDat[i];
		if (curveDat[i] > maxY)
			maxY = curveDat[i];

		//if (!strcmp(cName,"a2b0L1c1Curve")) {
	    //fprintf(stderr," %d ",ldat[i]);
		//  fprintf(stderr," %d ",(int) curveDat[i]);
		//}
			*/
	}

		
	//if (!strcmp(cName,"a2b0L1c1Curve")) {
	//  fprintf(stderr,"\n minX %f  maxX %f   minY %f  maxY %f  %d entries\n",minX,maxX,minY,maxY,siz);
	//}
		
	curve = (GtkCurve *) lookup_widget (widget, cName);

	gtk_widget_show ((GtkWidget *) widget);

	gtk_curve_set_range (curve, minX, maxX, minY, maxY);
	gtk_curve_set_vector (curve, (int) siz, curveDat);
	
	//gtk_curve_set_curve_type(curve,GTK_CURVE_TYPE_FREE);

	/*
	gtk_curve_get_vector(curve,(int) siz, curveDat);
	for (i = 0; i < siz; i++)
	{
	  fprintf(stderr,"read back i= %d cv= %d = %f   ldat= %d\n",i,(int) curveDat[i],curveDat[i],ldat[i]);
	}
	*/

	free (curveDat);

}

int
loadLCurve (LPLUT lut, GtkWidget * window, char *cName, int L, int channel)
{
  //fprintf(stderr,"loadlcurve: L= %d ",L);
	if (L == 1)
	{
		if (lut->InputChan > channel) {
			loadCurveLx (window, cName, lut->L1[channel],
				     lut->InputEntries);
			//setLcurveGamma(window, cName, lut->L1[channel],
			//	     lut->InputEntries);
		} else
			//hideWidget (window, cName);
			return FALSE;
	}
	else if (L == 2)
	{
		if (lut->OutputChan > channel) {
			loadCurveLx (window, cName, lut->L2[channel],
				     lut->OutputEntries);
			//setLcurveGamma(window, cName, lut->L2[channel],
			//	     lut->InputEntries);
		} else
			//hideWidget (window, cName);
			return FALSE;
	}
	else
	{
		fprintf (stderr, "LoadLCurve: L not 1 or 2 %d\n", L);
		return FALSE;
	}
	return TRUE;
}

void
loadABLcurves (LPLUT lut, GtkWidget * window, int a2b, char *dir)
{
	char text[16];
	int chan, L;

	for (L = 1; L < 3; L++)
	{
		for (chan = 0; chan < PICCEL_MAX_CHANNELS; chan++)
		{
			sprintf (text, "%s%dL%dc%dCurve", dir, a2b, L,
				 chan + 1);
			if (loadLCurve (lut, window, text, L, chan)) {
				sprintf(text,"%s%dL%dc%dvb",dir, a2b, L,
					 chan + 1);
				gtk_widget_show(lookup_widget(window,text));
			}
		}
	}
}

void
loadABCLUT (LPLUT lut, GtkWidget * window, int a2b, char *dir)
{
	char text[16];
	GtkSpinButton *sb;

	sprintf (text, "%s%dclutSB", dir, a2b);
	sb = (GtkSpinButton *) lookup_widget (window, text);
	gtk_spin_button_set_range (sb, (gdouble) 0,
				   (gdouble) lut->cLutPoints -1);

}

void
loadABEditTab (GtkWidget * window, int a2b, LPLUT lut, char *dir)
{
	char text[24];

	sprintf (text, "%s%dinputChan", dir, a2b);
	loadEntryWithUInt (window, text, lut->InputChan);
	sprintf (text, "%s%doutputChan", dir, a2b);
	loadEntryWithUInt (window, text, lut->OutputChan);
	sprintf (text, "%s%dinputEntries", dir, a2b);
	loadEntryWithUInt (window, text, lut->InputEntries);
	sprintf (text, "%s%doutputEntries", dir, a2b);
	loadEntryWithUInt (window, text, lut->OutputEntries);
	sprintf (text, "%s%dclutPoints", dir, a2b);
	loadEntryWithUInt (window, text, lut->cLutPoints);
	sprintf (text, "%s%dclutBytes", dir, a2b);
	loadEntryWithUInt (window, text, lut->Tsize);

	loadABLcurves (lut, window, a2b, dir);
	loadABCLUT (lut, window, a2b, dir);
}

/* Edit window A to B tab
*/

void
loadEditWinAtoBTab (char pid)
{

	LPLUT lut;
	GtkWidget *window;
	GtkWidget *vbox;

	if (!(window = pickEditWindow (pid)))
		return;

	if ((lut = getLUT (pid, icSigAToB0Tag)))
	{
		loadABEditTab (window, 0, lut, "a2b");
		cmsFreeLUT (lut);
	}
	else
	{
	  fprintf(stderr,"no a2b0 tag\n");
		vbox = lookup_widget (window, "A2B0vbox");
		gtk_widget_hide (vbox);
	}

	if ((lut = getLUT (pid, icSigAToB1Tag)))
	{
		loadABEditTab (window, 1, lut, "a2b");
		cmsFreeLUT (lut);
	}
	else
	{
	  fprintf(stderr,"no a2b1 tag\n");
		vbox = lookup_widget (window, "A2B1vbox");
		gtk_widget_hide (vbox);
	}

	if ((lut = getLUT (pid, icSigAToB2Tag)))
	{
		loadABEditTab (window, 2, lut, "a2b");
		cmsFreeLUT (lut);
	}
	else
	{
	  fprintf(stderr,"no a2b1 tag\n");
		vbox = lookup_widget (window, "A2B2vbox");
		gtk_widget_hide (vbox);
	}



	if ((lut = getLUT (pid, icSigBToA0Tag)))
	{
		loadABEditTab (window, 0, lut, "b2a");
		cmsFreeLUT (lut);
	}
	else
	{
	  fprintf(stderr,"no b2a0 tag\n");
		vbox = lookup_widget (window, "B2A0vbox");
		gtk_widget_hide (vbox);
	}

	if ((lut = getLUT (pid, icSigBToA1Tag)))
	{
		loadABEditTab (window, 1, lut, "b2a");
		cmsFreeLUT (lut);
	}
	else
	{
	  fprintf(stderr,"no b2a1 tag\n");
		vbox = lookup_widget (window, "B2A1vbox");
		gtk_widget_hide (vbox);
	}

	if ((lut = getLUT (pid, icSigBToA2Tag)))
	{
		loadABEditTab (window, 2, lut, "b2a");
		cmsFreeLUT (lut);
	}
	else
	{
	  fprintf(stderr,"no b2a2 tag\n");
		vbox = lookup_widget (window, "B2A2vbox");
		gtk_widget_hide (vbox);
	}

}


/* Edit Gamut support routines
*/
void
loadGamutLcurves (LPLUT lut, GtkWidget * window)
{
	char text[16];
	int chan, L;

	for (L = 1; L < 3; L++)
	{
		for (chan = 0; chan < PICCEL_MAX_CHANNELS; chan++)
		{
			sprintf (text, "gmutL%dc%dCurve", L, chan + 1);
			loadLCurve (lut, window, text, L, chan);
		}
	}
}

void
loadGamutCLUT (LPLUT lut, GtkWidget * window)
{
	GtkSpinButton *sb;

	sb = (GtkSpinButton *) lookup_widget (window, "gmutclutSB");
	gtk_spin_button_set_range (sb, (gdouble) 0,
				   (gdouble) lut->cLutPoints -1);


}

void
loadGamutEditTab (GtkWidget * window, LPLUT lut)
{
	loadEntryWithUInt (window, "gamutInputChan", lut->InputChan);
	loadEntryWithUInt (window, "gamutOutputChan", lut->OutputChan);
	loadEntryWithUInt (window, "gamutInputEntries", lut->InputEntries);
	loadEntryWithUInt (window, "gamutOutputEntries", lut->OutputEntries);
	loadEntryWithUInt (window, "gamutClutPoints", lut->cLutPoints);
	loadEntryWithUInt (window, "gamutClutBytes", lut->Tsize);

	loadGamutLcurves (lut, window);
	loadGamutCLUT (lut, window);
}

/* Edit window Gamut tab
*/

void
loadEditWinGamutTab (char pid)
{

	LPLUT lut;
	GtkWidget *window;
	GtkWidget *hbox;

	if (!(window = pickEditWindow (pid)))
		return;

	if ((lut = getLUT (pid, icSigGamutTag)))
	{
		loadGamutEditTab (window, lut);
		cmsFreeLUT (lut);
	}
	else
	{
		fprintf(stderr,"no gamut tag\n");
		hbox = lookup_widget (window, "gamutHbox");
		gtk_widget_hide (hbox);
	}
}

/* Edit window TRC tab
*/


void
setCurveGamma(GtkWidget * widget, char * sbname, unsigned short int *ldat, unsigned int siz) {
	double estG;
	GtkWidget *w;

	w = lookup_widget(widget,sbname);
	estG = cmsEstimateGammaEx(ldat,siz,0.7) * 100.0;
	
	//fprintf(stderr,"est gamma= %lf  sbname= %s\n",estG,sbname);
	
	SB_NO_Update = TRUE;

	if (w) 
	  gtk_spin_button_set_value((GtkSpinButton *) w,estG);
	
	SB_NO_Update = FALSE;
}

void 
loadTRCEditTab(GtkWidget *widget, LPGAMMATABLE gt, char *prefix) 
{
  char text[32];
  unsigned short int *ldat;
  unsigned int siz;
  GtkEntry *w;

  ldat = gt->GammaTable;
  siz = gt->nEntries;

  sprintf(text,"%sEntries",prefix);
  w = (GtkEntry *) lookup_widget(widget,text);
  sprintf(text,"%d",siz);
  gtk_entry_set_text(w, (const gchar *) text);

  sprintf(text,"%sCurve",prefix);
  loadCurveLx(widget, text, ldat, siz);
  
  sprintf(text,"%sVbox",prefix);
  gtk_widget_show(lookup_widget(widget,text));

  sprintf(text,"%sSb",prefix);
  setCurveGamma(widget,text,ldat,siz);
}

void
loadEditWinTRCTab (char pid)
{

	LPGAMMATABLE gt;
	GtkWidget *window;
	GtkWidget *hbox;

	if (!(window = pickEditWindow (pid)))
		return;

	if ((gt = getGT (pid, icSigRedTRCTag)))
	{
		loadTRCEditTab (window, gt,"trcRed");
		cmsFreeGamma (gt);
	}
	else
	{
		fprintf(stderr,"no TRCRed tag\n");
		hbox = lookup_widget (window, "trcRedHbox");
		gtk_widget_hide (hbox);
	}

	if ((gt = getGT (pid, icSigGreenTRCTag)))
	{
		loadTRCEditTab (window, gt,"trcGrn");
		cmsFreeGamma (gt);
	}
	else
	{
		fprintf(stderr,"no TRCGreen tag\n");
		hbox = lookup_widget (window, "trcGrnHbox");
		gtk_widget_hide (hbox);
	}

	if ((gt = getGT (pid, icSigBlueTRCTag)))
	{
		loadTRCEditTab (window, gt,"trcBlu");
		cmsFreeGamma(gt);
	}
	else
	{
		fprintf(stderr,"no TRCBlue tag\n");
		hbox = lookup_widget (window, "trcBluHbox");
		gtk_widget_hide (hbox);
	}

	if ((gt = getGT (pid, icSigGrayTRCTag)))
	{
		loadTRCEditTab (window, gt,"trcGry");
		cmsFreeGamma (gt);
	}
	else
	{
		fprintf(stderr,"no TRCGray tag\n");
		hbox = lookup_widget (window, "trcGryHbox");
		gtk_widget_hide (hbox);
	}

}




/* Edit window Preview tab
*/

void
loadEditWinPreviewTab (char pid)
{

	LPLUT lut;
	GtkWidget *window;
	GtkWidget *hbox;

	if (!(window = pickEditWindow (pid)))
		return;

	if ((lut = getLUT (pid, icSigPreview0Tag)))
	{
		loadABEditTab (window, 0, lut, "prv");
		cmsFreeLUT (lut);
	}
	else
	{
		fprintf(stderr,"no Preview0 tag\n");
		hbox = lookup_widget (window, "prv0Hbox");
		gtk_widget_hide (hbox);
	}

	if ((lut = getLUT (pid, icSigPreview1Tag)))
	{
		loadABEditTab (window, 1, lut, "prv");
		cmsFreeLUT (lut);
	}
	else
	{
		fprintf(stderr,"no Preview1 tag\n");
		hbox = lookup_widget (window, "prv1Hbox");
		gtk_widget_hide (hbox);
	}

	if ((lut = getLUT (pid, icSigPreview2Tag)))
	{
		loadABEditTab (window, 2, lut, "prv");
		cmsFreeLUT (lut);
	}
	else
	{
		fprintf(stderr,"no Preview2 tag\n");
		hbox = lookup_widget (window, "prv2Hbox");
		gtk_widget_hide (hbox);
	}
}



/* CLUT rendering support routines
*/
char
getClutWidgetPID (GtkWidget * widget)
{
	GtkWidget *window;
	char *title;

	window = gtk_widget_get_toplevel (widget);
	title = (char *) gtk_window_get_title (GTK_WINDOW (window));
	//fprintf(stderr,"a2b0clutDrawArea expose  window= %s\n",title);

	if (strstr (title, "input") != NULL)
		return 'i';
	else if (strstr (title, "output") != NULL)
		return 'o';
	else if (strstr (title, "monitor") != NULL)
		return 'm';
	else
	{
		fprintf (stderr,
			 "getClutWidgetPID -- could not identify toplevel window title %s\n",
			 title);
		return '\0';
	}
}

void
clearClutSlice (GtkWidget * drawArea)
{
	gdk_draw_rectangle (drawArea->window,
			    drawArea->style->
			    bg_gc[GTK_WIDGET_STATE (drawArea)], TRUE, 0, 0,
			    drawArea->allocation.width,
			    drawArea->allocation.height);
}


void
renderClutSlice (GtkWidget * drawArea, unsigned int clutPoints,
		 unsigned int Tsize, LPWORD T, int slice, icColorSpaceSignature sig) {

  int sliceSize = 0;
  int i, j;
  int side;
  int x, y;
  
  int interCellSpace = 1;
  int cellSide;
  int cellStep;
  
  int Tunit = 3;		// 3 dwords = 6 bytes
  LPWORD Tbase;
  
  GdkColor gcolor;
  GdkGC *gc;
  double color[3];
  
  Tbase = T + (slice * (clutPoints * clutPoints * Tunit));
  
  if (Tbase > (T + Tsize)) {
    fprintf (stderr, "overshot T: %d  %lx > %lx \n", slice,
	     (unsigned long) Tbase, (unsigned long) T);
    return;
  }
  
  gc = gdk_gc_new (drawArea->window);
  sliceSize = (int) clutPoints *clutPoints;
  side = drawArea->allocation.width;
  if (side > drawArea->allocation.height)
    side = drawArea->allocation.height;
  cellSide = (side - (interCellSpace * clutPoints)) / clutPoints;
  cellStep = cellSide + interCellSpace;
  
  clearClutSlice (drawArea);
  
  for (i = 0; i < clutPoints; i++) {
    y = i * cellStep;
    for (j = 0; j < clutPoints; j++) {
      x = j * cellStep;
      
      color[0] = *Tbase++;      // L or X
      color[1] = *Tbase++;      // a or Y
      color[2] = *Tbase++;      // b or Z
	    
      //fprintf(stderr,"raw: r %d g %d b %d -- ",gcolor.red,gcolor.green,gcolor.blue);
      if (sig == icSigLabData) {
	//lab2xyzA(color);
	//fprintf(stderr,"lab2xyz result: r %d g %d b %d -- ",gcolor.red,gcolor.green,gcolor.blue);
      }

      //xyz2rgbA(color,2.4);

      //fprintf(stderr,"xyz2rgb result: r %d g %d b %d \n",gcolor.red,gcolor.green,gcolor.blue);
	    
      /*
       * fprintf(stderr,"finish: r %d g %d b %d  x %d  y %d slice %d  i %d  j %d  Tbase %lx > T+Tsize %lx \n",
       * gcolor.red,gcolor.green,gcolor.blue,x,y,slice,i,j,(unsigned long) Tbase, (unsigned long) (T + Tsize) );
       */
      
      gcolor.pixel = 0;
      gcolor.red   = (guint16)  round( color[0] );
      gcolor.green = (guint16)  round( color[1] );
      gcolor.blue  = (guint16)  round( color[2] );

      gdk_gc_set_rgb_fg_color (gc, &gcolor);	    
      gdk_draw_rectangle (drawArea->window, gc, TRUE, x, y, cellSide, cellSide);
      
    }
  }
}

icTagSignature
getTag (char a2b, int i, char *str) {
  switch (a2b) {
  case 'a':
    switch (i) {
    case 0:
      return icSigAToB0Tag;
      break;
    case 1:
      return icSigAToB1Tag;
      break;
    case 2:
      return icSigAToB2Tag;
      break;
    default:
      fprintf(stderr,"failed to set tag %c %d\n",a2b,i);
      break;
    }
    break;
  case 'b':
    switch (i) {
    case 0:
      return icSigBToA0Tag;
      break;
    case 1:
      return icSigBToA1Tag;
      break;
    case 2:
      return icSigBToA2Tag;
      break;
    default:
      fprintf(stderr,"failed to set tag %c %d\n",a2b,i);
      break;
    }
    break;
  case 'g':    // gamut
    return icSigGamutTag;
    break;
  case 't':    // trc
    switch(str[5]) {
    case 'd':
      return icSigRedTRCTag;
      break;
    case 'n':
      return icSigGreenTRCTag;
      break;
    case 'u':
      return icSigBlueTRCTag;
      break;
    case 'y':
      return icSigGrayTRCTag;
      break;
    default:
      fprintf(stderr,"failed to set tag %c %c\n",a2b,str[5]);
      break;
    }		
    break;
  default:
    fprintf(stderr,"failed to set tag %c %d\n",a2b,i);
    break;
  }
	return 0;
}

void
drawClut (char pid, GtkWidget * drawArea, char *dir, int a2b)
{
	icTagSignature tag = 0;
	LPLUT lut;
	char text[16];
	int slice = 0;

	if (!strcmp (dir, "a2b"))
	{
		sprintf (text, "a2b%dclutSB", a2b);
		tag = getTag('a',a2b,NULL);
	}
	else if (!strcmp (dir, "b2a"))
	{
		sprintf (text, "b2a%dclutSB", a2b);
		tag = getTag('b',a2b,NULL);
	}
	else if (!strcmp (dir, "gamut"))
	{
		tag = icSigGamutTag;
		sprintf (text, "gmutclutSB");
	}
	else if (!strcmp (dir, "prv0"))
	{
		tag = icSigPreview0Tag;
		sprintf (text, "prv0clutSB");
	}
	else if (!strcmp (dir, "prv1"))
	{
		tag = icSigPreview0Tag;
		sprintf (text, "prv1clutSB");
	}
	else if (!strcmp (dir, "prv2"))
	{
		tag = icSigPreview0Tag;
		sprintf (text, "prv2clutSB");
	}
	else
	{
		fprintf (stderr, "drawClut: unhandled clut class %s\n", dir);
		return;
	}

	slice = ((int)
		 gtk_spin_button_get_value ((GtkSpinButton *)
					    lookup_widget (drawArea,
							   text)));
	//fprintf(stderr,"drawClut: slice %d\n",slice);

	if (!(lut = getLUT (pid, tag)))
	{
		fprintf (stderr,
			 "drawClut: failed to get LUT for %s %d (%c)\n", dir,
			 a2b, pid);
		return;
	}

	renderClutSlice (drawArea, lut->cLutPoints, lut->Tsize, lut->T, slice, getPCS(pid));

	cmsFreeLUT (lut);
}

gboolean
handleClutExpose (GtkWidget * widget, char *dir, int a2b)
{
	char pid;

	if ((pid = getClutWidgetPID (widget)) == '\0')
		return FALSE;

	drawClut (pid, widget, dir, a2b);

	return FALSE;
}

/* CLUT expose
*/
gboolean
on_a2b0clutDrawArea_expose_event (GtkWidget * widget,
				  GdkEventExpose * event, gpointer user_data)
{
	return handleClutExpose (widget, "a2b", 0);
}

gboolean
on_a2b1clutDrawArea_expose_event       (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{

	return handleClutExpose (widget, "a2b", 1);
}

gboolean
on_a2b2clutDrawArea_expose_event       (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{

	return handleClutExpose (widget, "a2b", 2);
}

gboolean
on_b2a0clutDrawArea_expose_event       (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{

	return handleClutExpose (widget, "b2a", 0);
}

gboolean
on_b2a1clutDrawArea_expose_event       (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{

	return handleClutExpose (widget, "b2a", 1);
}


gboolean
on_b2a2clutDrawArea_expose_event       (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{

	return handleClutExpose (widget, "b2a", 2);
}



/* CLUT spinbutton value changed
*/
void
on_a2b0clutSB_value_changed (GtkSpinButton * spinbutton, gpointer user_data)
{
	handleClutExpose (lookup_widget
			  (GTK_WIDGET (spinbutton), "a2b0clutDrawArea"),
			  "a2b", 0);
}

void
on_a2b1clutSB_value_changed            (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
	handleClutExpose (lookup_widget
			  (GTK_WIDGET (spinbutton), "a2b1clutDrawArea"),
			  "a2b", 1);
}


void
on_a2b2clutSB_value_changed            (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
	handleClutExpose (lookup_widget
			  (GTK_WIDGET (spinbutton), "a2b2clutDrawArea"),
			  "a2b", 2);

}


void
on_b2a0clutSB_value_changed            (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
	handleClutExpose (lookup_widget
			  (GTK_WIDGET (spinbutton), "b2a0clutDrawArea"),
			  "b2a", 0);

}


void
on_b2a1clutSB_value_changed            (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
	handleClutExpose (lookup_widget
			  (GTK_WIDGET (spinbutton), "b2a1clutDrawArea"),
			  "b2a", 1);

}


void
on_b2a2clutSB_value_changed            (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
	handleClutExpose (lookup_widget
			  (GTK_WIDGET (spinbutton), "b2a2clutDrawArea"),
			  "b2a", 2);

}


/* CLUT rebuild button
*/

static void genClutWin(GtkButton *button, char *base) {

  char wintitle[48];
  char prof[12];

  GtkWidget *widget = (GtkWidget *) button;
  GtkWidget *window = gtk_widget_get_toplevel(widget);
  char *parent_title = (char *) gtk_window_get_title(GTK_WINDOW( window )); 

  sscanf(parent_title,"piccel edit %s profile",&prof);
  sprintf(wintitle,"%s profile %s CLUT generator",prof,base);

  piccelGenCLUTWindow = (GtkWindow *) create_piccel_gen_clut();
  gtk_window_set_title(piccelGenCLUTWindow, (const gchar *) wintitle);
  gtk_widget_show ( (GtkWidget *) piccelGenCLUTWindow );
}

void
on_a2b0clutRebuildButton_clicked (GtkButton * button, gpointer user_data)
{
  genClutWin(button,"a2b0");
}

void
on_a2b1clutRebuildButton_clicked       (GtkButton       *button,
                                         gpointer         user_data)
{
  genClutWin(button,"a2b1");
}

void
on_a2b2clutRebuildButton_clicked       (GtkButton       *button,
                                        gpointer         user_data)
{
  genClutWin(button,"a2b2");
}


void
on_b2a0clutRebuildButton_clicked       (GtkButton       *button,
                                         gpointer         user_data)
{
  genClutWin(button,"b2a0");
}


void
on_b2a1clutRebuildButton_clicked       (GtkButton       *button,
                                         gpointer         user_data)
{
  genClutWin(button,"b2a1");
}


void
on_b2a2clutRebuildButton_clicked       (GtkButton       *button,
                                         gpointer         user_data)
{
  genClutWin(button,"b2a2");
}

/** Edit window main actions
*/

void 
setCharTargetData(cmsHPROFILE profile, GtkWidget *widget) {
  GtkTextBuffer *tbuff;
  GtkTextIter start,end;

  tbuff = gtk_text_view_get_buffer((GtkTextView *) lookup_widget(widget,"charTargTextView"));
  if (gtk_text_buffer_get_char_count(tbuff) == 0) return;
  gtk_text_buffer_get_bounds(tbuff,&start,&end);
  cmsAddTag(profile,
	    icSigCharTargetTag,
	    (void*) gtk_text_buffer_get_text(tbuff,&start,&end, FALSE));
}


void 
setColorants(cmsHPROFILE hprofile, GtkWidget *widget) {
  double col[3];
  
  if (readEntryDouble(widget,"redColXEntry",&(col[0]))
      &&
      readEntryDouble(widget,"redColYEntry",&(col[1]))
      &&
      readEntryDouble(widget,"redColZEntry",&(col[2])) ) {
    setColorant(hprofile,icSigRedColorantTag,col);
  }

  if (readEntryDouble(widget,"greenColXEntry",&(col[0]))
      &&
      readEntryDouble(widget,"greenColYEntry",&(col[1]))
      &&
      readEntryDouble(widget,"greenColZEntry",&(col[2])) ) {
    setColorant(hprofile,icSigGreenColorantTag,col);
  }

  if (readEntryDouble(widget,"blueColXEntry",&(col[0]))
      &&
      readEntryDouble(widget,"blueColYEntry",&(col[1]))
      &&
      readEntryDouble(widget,"blueColZEntry",&(col[2])) ) {
    setColorant(hprofile,icSigBlueColorantTag,col);
  }

}

void
setMediaPoints(cmsHPROFILE hprofile, GtkWidget *widget) {
  double mpt[3];
  
  if (readEntryDouble(widget,"mWPXEntry",&(mpt[0]))
      &&
      readEntryDouble(widget,"mWPYEntry",&(mpt[1]))
      &&
      readEntryDouble(widget,"mWPZEntry",&(mpt[2])) ) {
    setMediaPt(hprofile,icSigMediaWhitePointTag,mpt);
  }

  if (readEntryDouble(widget,"mBPXEntry",&(mpt[0]))
      &&
      readEntryDouble(widget,"mBPYEntry",&(mpt[1]))
      &&
      readEntryDouble(widget,"mBPZEntry",&(mpt[2])) ) {
    setMediaPt(hprofile,icSigMediaBlackPointTag,mpt);
  }

  
}

void
setTRC(UL flags, cmsHPROFILE outprofile, char *wname, char pid, icTagSignature tag, GtkWidget *widget) {

  LPGAMMATABLE gt;
  
  if ((gt = getGT(pid,tag))) {
    if (flags) {
      // use modified data
      int numEntries,i;
      gfloat *curveDat;

      fprintf(stderr,"updating TRC gama curve\n");
      numEntries = gt->nEntries;
      cmsFreeGamma(gt);
      gt = cmsAllocGamma(numEntries);
      if (!(curveDat = malloc (numEntries * sizeof (float))))
      {
	fprintf (stderr, "unable to allocate memory to read TRC gamma curve %s\n",wname);
	return;
      }
      gtk_curve_get_vector((GtkCurve*) lookup_widget(widget,wname), numEntries, curveDat);
      for (i=0;i<numEntries;i++) {
	gt->GammaTable[i] = (WORD) curveDat[i];
      }
      free(curveDat);
      
    }  // else use existing profile gamma table, already loaded.

    cmsAddTag(outprofile,tag,(void*) gt);
    cmsFreeGamma(gt);
  }
}

void 
setLinearTable(LPLUT newlut, UL w2Flags, GtkWidget *widget, char *base, int ltid) {
  
  int ch=0;
  int entries;
  int channels;
  WORD *gammatable;

  switch (ltid) {
  case 1:
    entries = newlut->InputEntries;
    channels = newlut->InputChan;
    break;
  case 2:
    entries = newlut->OutputEntries;
    channels = newlut->OutputChan;
    break;
  default:
    fprintf(stderr,"setLinearTable not L1 or L2");
    exit(-1);
    break;
  }

  if ( w2Flags ) {
    // yes, an Lx curve has been modified so read data from widget

    for (ch=0;ch<channels;ch++) {
      UL ch_select=0;
      switch (ch) {
      case 0 : ch_select = ABG_C1_CURVES;  break;
      case 1 : ch_select = ABG_C2_CURVES;  break;
      case 2 : ch_select = ABG_C3_CURVES;  break;
      case 3 : ch_select = ABG_C4_CURVES;  break;
      default:
	fprintf(stderr,"more than 4 input channels, keep working!");
	assert(PICCEL_MAX_CHANNELS <= 4);
	exit(-1);
      }

      if (w2Flags & ch_select) {
	char wID[16];
	GtkCurve *cwidget;
	gfloat *cvector;
	int i;
	
	sprintf(wID,"%sL%dc%dCurve",base,ltid,ch+1);

	fprintf(stderr,"output for curve %s \n",wID);

	cwidget = (GtkCurve *) lookup_widget(widget,wID);
	cvector = malloc(sizeof(float) * entries);
	gtk_curve_get_vector(cwidget,entries,cvector);

	// gtk_curve_get doesn't give back what was put in !
	
	//for (i=0; i< entries; i++) {
	//  fprintf(stderr," %d ",(int) cvector[i]);
	//}
	//fprintf(stderr,"\n %d entries \n",entries);
	
	//gammatable = (ltid == 1 ? ((LPGAMMATABLE) newlut->L1[ch])->GammaTable : ((LPGAMMATABLE) newlut->L2[ch])->GammaTable );
	gammatable = (ltid == 1 ? newlut->L1[ch] : newlut->L2[ch] );
	
	//fprintf(stderr,"setlineartable entries= %d ltid= %d\n",entries,ltid);
	for (i=0;i<entries;i++) {
	  //fprintf(stderr,"i= %d gt= %d cv= %d = %f\n",i,(int) gammatable[i], (WORD) cvector[i], cvector[i]);
	  gammatable[i] = (WORD) cvector[i]; 
	}

	free (cvector);

      } 
    } // for channels
  }

}


static
LPVOID DupBlockTab(LPVOID Org, size_t size)   // direct copy from lcms/cmslut.c
{
    LPVOID mem = malloc(size);

    CopyMemory(mem, Org, size);
    return mem;
}


void
setLUT(UL topflag, UL flags, cmsHPROFILE outProfile, char *base, char pid, icTagSignature tag, GtkWidget *widget) {

  LPLUT lut=0;
  UL workFlags=0;

  if (! (lut = getLUT(pid,tag))) {  // easiest to start with what is already there and modify
    fprintf(stderr,"setLUT %s failed to open current %c lut\n",base,pid);
    return;
  }

  //lut = getLUT(pid,tag);   // easiest to start with what is already there and modify

  if (workFlags = (topflag & flags)) {

  // part of lut is modified

    unsigned int InputChan;
    unsigned int OutputChan;
    unsigned int InputEntries;
    unsigned int OutputEntries;
    unsigned int cLutPoints;
    char wID[24];


    fprintf(stderr,"modified lut, base= %s\n",base);

    sprintf(wID,"%sinputChan",base);
    sscanf(gtk_entry_get_text((GtkEntry*) lookup_widget(widget,wID)),"%d",&InputChan);
    sprintf(wID,"%soutputChan",base);
    sscanf(gtk_entry_get_text((GtkEntry*) lookup_widget(widget,wID)),"%d",&OutputChan);
    sprintf(wID,"%sinputEntries",base);
    sscanf(gtk_entry_get_text((GtkEntry*) lookup_widget(widget,wID)),"%d",&InputEntries);
    sprintf(wID,"%soutputEntries",base);
    sscanf(gtk_entry_get_text((GtkEntry*) lookup_widget(widget,wID)),"%d",&OutputEntries);
    sprintf(wID,"%sclutPoints",base);
    sscanf(gtk_entry_get_text((GtkEntry*) lookup_widget(widget,wID)),"%d",&cLutPoints);


    /*
      and working here ...
    */

    // CLUT
    // do this before Lx curves as alloc3Dgrid sets input/outputChan
    //fprintf(stderr,"test save clut: logic %lx wl %lx\n",(flags & ( topflag & ABG_CLUTS )),workLUT );

    if ((flags & ( topflag & ABG_CLUTS )) && (workLUT != (LPLUT) 0) && tag == workTag) {
    // hack!!!!!!!!!! to get 3 luts updated
    //if ((flags & ( topflag & ABG_CLUTS )) && (workLUT != (LPLUT) 0)) {
      int i;
      int count;
      LPWORD T1,T2;

      fprintf(stderr,"copying new clut data %s!\n",base); 
      // free existing 3d grid?
      //cmsAlloc3DGrid(lut,cLutPoints,InputChan,OutputChan);
      T1 = lut->T;
      T2 = workLUT->T;
      count = lut->Tsize / sizeof(WORD);

      // populate with data from widget 

      for (i=0;i<count;i++) {
	*T1++ = *T2++;
      }

    } 

    // L1 curves
    setLinearTable(lut,(workFlags & ABG_L1_CURVES),widget,base,1);
    // L2 curves
    setLinearTable(lut,(workFlags & ABG_L2_CURVES),widget,base,2);

  } else {
    fprintf(stderr,"setLUT %s %c copying unmodified lut\n",base,pid);
  }

  cmsAddTag(outProfile, tag, (void *) lut);  // add to output profile whether modified or not
  //cmsFreeLUT(lut);

}

void
writeNewProfile(char *filename, GtkWidget * widget) {

  cmsHPROFILE outProfile;
  char pid='\0';
  BOOL rslt;

  pid = getClutWidgetPID(widget);

  if ( (outProfile = cmsOpenProfileFromFile(filename,"w")) == (cmsHPROFILE) NULL ) {
    sprintf(errstr,"unable to open %s for writing profile\n",filename);
    alert();
    return;
  }


  cmsSetDeviceClass(outProfile, convertDeviceClassCB( gtk_combo_box_get_active((GtkComboBox*) lookup_widget(widget,"devClassCB") ) ) );
  cmsSetColorSpace(outProfile, convertColSpaceCB( gtk_combo_box_get_active((GtkComboBox*) lookup_widget(widget,"colSpaceCB") ) ) );
  cmsSetPCS(outProfile, convertColSpaceCB( gtk_combo_box_get_active((GtkComboBox*) lookup_widget(widget,"pcsCB") ) ) );
  cmsSetRenderingIntent(outProfile, convertIntentCB( gtk_combo_box_get_active((GtkComboBox*) lookup_widget(widget,"editIntentCB") ) ) );

  setCharTargetData(outProfile,widget);

  cmsAddTag(outProfile,icSigCopyrightTag,(void*) gtk_entry_get_text((GtkEntry*) lookup_widget(widget,"copyrightEntry")));
  cmsAddTag(outProfile,icSigProfileDescriptionTag,(void*) gtk_entry_get_text((GtkEntry*) lookup_widget(widget,"profDescEntry")));
  cmsAddTag(outProfile,icSigDeviceMfgDescTag,(void*) gtk_entry_get_text((GtkEntry*) lookup_widget(widget,"devMfgEntry")));
  cmsAddTag(outProfile,icSigDeviceModelDescTag,(void*) gtk_entry_get_text((GtkEntry*) lookup_widget(widget,"devModelEntry")));



  setColorants(outProfile,widget);

  setMediaPoints(outProfile,widget);

  setTRC(TRCRED & update_flags2,outProfile,"trcRedCurve",pid, icSigRedTRCTag,widget);
  setTRC(TRCGRN & update_flags2,outProfile,"trcGrnCurve",pid, icSigGreenTRCTag,widget);
  setTRC(TRCBLU & update_flags2,outProfile,"trcBluCurve",pid, icSigBlueTRCTag,widget);
  setTRC(TRCGRY & update_flags2,outProfile,"trcGryCurve",pid, icSigGrayTRCTag,widget);


  setLUT(A2B0, update_flags1,outProfile,"a2b0",pid, icSigAToB0Tag,widget);
  setLUT(A2B1, update_flags1,outProfile,"a2b1",pid, icSigAToB1Tag,widget);
  setLUT(A2B2, update_flags1,outProfile,"a2b2",pid, icSigAToB2Tag,widget);

  setLUT(B2A0, update_flags1,outProfile,"b2a0",pid, icSigBToA0Tag,widget);
  setLUT(B2A1, update_flags1,outProfile,"b2a1",pid, icSigBToA1Tag,widget);
  setLUT(B2A2, update_flags1,outProfile,"b2a2",pid, icSigBToA2Tag,widget);

  setLUT(GMUT, update_flags1,outProfile,"gmut",pid, icSigGamutTag,widget);

  // untested...
  setLUT(PRV0, update_flags1,outProfile,"prv0",pid, icSigPreview0Tag,widget);
  setLUT(PRV1, update_flags1,outProfile,"prv1",pid, icSigPreview1Tag,widget);
  setLUT(PRV2, update_flags1,outProfile,"prv2",pid, icSigPreview2Tag,widget);
  //
  rslt = cmsCloseProfile(outProfile);
  fprintf(stderr,"wrote profile %s result= %d\n",filename,rslt);


  // obscure problem when overwriting profile with new data and lut not modified: in memory version loses lut ?
  switch (pid) {
  case 'o':  
    if (!strcmp(oprof,filename)) {
      fprintf(stderr,"reloading output profile %s\n",oprof);
      pLoadOutputProfile(oprof);
    }
    break;
  case 'i':
    if (!strcmp(iprof,filename)) {
      fprintf(stderr,"reloading input profile %s\n",iprof);
      pLoadInputProfile(iprof);
    }
    break;
  case 'm':
    if (!strcmp(mprof,filename)) {
      fprintf(stderr,"reloading monitor profile %s\n",mprof);
      pLoadMonitorProfile(mprof);
    }
    break;
  }



}

void
on_editUpdateButton_clicked (GtkButton * button, gpointer user_data)
{
  GtkWidget *widget = (GtkWidget *) button;
  GtkWidget *window = gtk_widget_get_toplevel(widget);
  char *parent_title = (char *) gtk_window_get_title(GTK_WINDOW( window )); 
  char prof[16];
  char *profilename = NULL;
  char *tmpfile="/tmp/piccel.icc";

  sscanf(parent_title,"piccel edit %s profile",prof);

  fprintf(stderr,"update: parent title= .%s. prifile= .%s.\n",parent_title,prof);
  
  writeNewProfile(tmpfile,(GtkWidget*) button);
  
  switch (prof[0]) {
  case 'i' :
    profilename = pLoadInputProfile (tmpfile);
    //loadEditWinAtoBTab ('i');
    break;
  case 'o' :
    profilename = pLoadOutputProfile (tmpfile);
    //loadEditWinAtoBTab ('o');
    break;
  case 'm' :
    profilename = pLoadMonitorProfile (tmpfile);
    //loadEditWinAtoBTab ('m');
    break;
  }
  
  
  if (profilename) {
    printf ("updated %s profile  .%s.\n",prof, profilename);
    setProfileName (prof[0], profilename,
		    GTK_WIDGET (button));
  } else {
    printf ("updated %s profile  dynamic\n",prof);
    setProfileName (prof[0], "dynamic",
		    GTK_WIDGET (button));
  }

  on_image_update_clicked(main_win_edit_btn,user_data);
}


void
on_editSaveButton_clicked (GtkButton * button, gpointer user_data)
{
  GtkWidget *fsdlg;

  fsdlg = create_filesaverdialog();
  setdir(fsdlg);
  if (gtk_dialog_run (GTK_DIALOG (fsdlg)) == GTK_RESPONSE_OK) {
    char *filename;
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(fsdlg));
    savedir(filename);
    writeNewProfile(filename,(GtkWidget*) button);
  } else {
    printf ("edit save: response not OK\n");
  }

  gtk_widget_destroy (fsdlg);

}

gboolean
on_piccel_edit_window2_destroy_event (GtkWidget * widget,
				      GdkEvent * event, gpointer user_data)
{
	return FALSE;
}


void
on_piccel_edit_window2_destroy (GtkObject * object, gpointer user_data)
{
	char *title;

	title = (char *) gtk_window_get_title (GTK_WINDOW (object));

	if (strstr (title, "input") != NULL)
		piccelEditIPWindow = 0;
	else if (strstr (title, "output") != NULL)
		piccelEditOPWindow = 0;
	else if (strstr (title, "monitor") != NULL)
		piccelEditMPWindow = 0;
	else
		fprintf (stderr,
			 "could not identify destroyed edit window - title %s\n",
			 title);

	fprintf (stderr, "closing edit window %s\n", title);


}

/*
* Generalised logistic curve from 
* http://en.wikipedia.org/wiki/Generalised_logistic_curve
*
*  Y = A+( C / (1+Texp(-B(X-M))^(1/T)) )
*
*  A = lower asymptote
*  C = upper asymptote minus A
*  M = time (X) of maximum growth (rate)
*  B = growth rate
*  T = affects near which asymptote maximum growth occurs
*/

GAMMATABLE *
buildTable(int numEntries,double p) {
	LPGAMMATABLE gt;
	int i;
	gt = cmsAllocGamma(numEntries);
	for (i=0;i<numEntries;i++)
		gt->GammaTable[i] = (int) (p * (65536.0 / (1+exp(-(0.005*(((double)i)- (numEntries/2.0)))))))
				+ 
				(((1.0-p)) * ((65536.0/numEntries) * i));
	// force the endpoints:
	gt->GammaTable[0] = 0;
	gt->GammaTable[numEntries-1] = 65535;
	return gt;
}

void
on_Lcurve_sb_value_changed             (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
	LPGAMMATABLE gt;
	char pid;
	char *wname;
	icTagSignature tag = 0;
	LPLUT lut;

	wname = (char *) gtk_widget_get_name((GtkWidget *) spinbutton);
	//fprintf(stderr,"sb val changed widget name = %s\n",wname);
	tag = getTag(wname[0],(wname[3]-'0'),wname);

	if ((pid = getClutWidgetPID ((GtkWidget *) spinbutton)) == '\0') 
		return;
	if ((lut = getLUT(pid,tag))) {
	  char cName[32];
	  int entries;
	  switch(wname[5]) {
		  case '1':
			  entries = lut->InputEntries;
		  break;
		  case '2':
			  entries = lut->OutputEntries;
		  break;
		  default:
		  	fprintf(stderr,"lc sb val changed not L1 or L2 : %c char 5 of  %s\n",wname[5],wname);
		    return;
		  break;
	  }
	  sprintf(cName,"%c2%c%cL%cc%cCurve",wname[0],wname[2],wname[3],wname[5],wname[7]);
	  //fprintf(stderr,"buildTable with %d entries %s\n",entries,cName);
	  gt = buildTable(entries,gtk_spin_button_get_value(spinbutton)/100.0);
	  cmsFreeLUT(lut);
	  loadCurveLx((GtkWidget *) spinbutton, cName, gt->GammaTable, gt->nEntries);
	  cmsFreeGamma(gt);
	}
	set_update_flags(wname);
}


void
on_LC_use_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
	char pid;
	char *wname;
	icTagSignature tag = 0;
	LPLUT lut;
	  char cName[32];


	wname = (char *) gtk_widget_get_name((GtkWidget *) button);
	fprintf(stderr,"LC use: widget name = %s\n",wname);
	tag = getTag(wname[0],(wname[3]-'0'),wname);
	  sprintf(cName,"%c2%c%cL%cc%cCurve",wname[0],wname[2],wname[3],wname[5],wname[7]);

	if ((pid = getClutWidgetPID ((GtkWidget *) button)) == '\0') 
		return;
	if ((lut = getLUT(pid,tag))) {
		GtkWidget *w;
		gfloat *vec;
		if ((vec = (gfloat *) malloc(lut->InputEntries * sizeof(gfloat))) 
			== (gfloat *) 0) {
				fprintf(stderr,"unable to allocate Lcurve vector space\n");
				return;
			}
		w = lookup_widget((GtkWidget*)button,cName);
		if (w) {
			int i;
			WORD *Lc;
			int chan;
			cmsHPROFILE pro;
			
			chan = wname[7];
			Lc = (WORD *) (wname[5]=='1' ? &(lut->L1[chan]) : &(lut->L2[chan]));
			fprintf(stderr,"get vector %d entries\n", lut->InputEntries);
			gtk_curve_get_vector((GtkCurve*)w,lut->InputEntries,vec);
			fprintf(stderr,"have vector, copy over Lc=%0x\n",(unsigned int)Lc);
			for (i=0;i<lut->InputEntries;i++) {
				fprintf(stderr,"%d/%d %d\n",i,lut->InputEntries, (unsigned short int) vec[i]);
				Lc[i] = (unsigned short int) vec[i];
			}
			fprintf(stderr,"finished copy\n");
			pro = chooseProfile(pid);
			fprintf(stderr,"setting L curve pro= %0x\n",(unsigned int) pro);
			//cmsAddTag(pro,tag,lut);
			fprintf(stderr,"finished add tag\n");
			
		}
		//fprintf(stderr,"free vec...\n");
		//free(vec);
		//fprintf(stderr,"free lut...\n");
		//cmsFreeLUT(lut);
		fprintf(stderr,"done\n");
	}
		
}

void
on_gmutclutSB_value_changed            (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
	handleClutExpose (lookup_widget
			  (GTK_WIDGET (spinbutton), "gmutclutDrawArea"),
			  "gamut", 0);
}


void
on_prv0clutSB_value_changed            (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
	handleClutExpose (lookup_widget
			  (GTK_WIDGET (spinbutton), "prv0clutDrawArea"),
			  "prv0", 0);

}


void
on_prv1clutSB_value_changed            (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
	handleClutExpose (lookup_widget
			  (GTK_WIDGET (spinbutton), "prv1clutDrawArea"),
			  "prv1", 0);

}


void
on_prv2clutSB_value_changed            (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
	handleClutExpose (lookup_widget
			  (GTK_WIDGET (spinbutton), "prv2clutDrawArea"),
			  "prv2", 0);

}


void
on_prv2clutRebuildButton_clicked       (GtkButton       *button,
                                        gpointer         user_data)
{

}

gboolean
on_gmutClutDrawArea_expose_event       (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{

	return handleClutExpose (widget, "gamut", 1);

  return FALSE;
}

void
on_TRCcurve_sb_value_changed           (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
	LPGAMMATABLE bgt,gt;
	char pid;
	char *wname;
	icTagSignature tag = 0;

	if (SB_NO_Update) return;

	wname = (char *) gtk_widget_get_name((GtkWidget *) spinbutton);
	//fprintf(stderr,"TRC sb val changed widget name = %s\n",wname);
	tag = getTag(wname[0],(wname[3]-'0'),wname);

	if ((pid = getClutWidgetPID ((GtkWidget *) spinbutton)) == '\0') 
		return;
	if ((gt = getGT(pid,tag))) {
	  char cName[32];
	  sprintf(cName,"trc%c%c%cCurve",wname[3],wname[4],wname[5]);
	  bgt = cmsBuildGamma(gt->nEntries,
		(double) (gtk_spin_button_get_value(spinbutton)/100.0));
	  cmsFreeGamma(gt);
	  loadCurveLx((GtkWidget *) spinbutton, cName, bgt->GammaTable, bgt->nEntries);
	  cmsFreeGamma(bgt);
	}
	
	set_update_flags(wname);

}



gboolean
on_clutDrawArea_button_press_event         (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{

	set_update_flags((char*) gtk_widget_get_name(widget));
	return FALSE;
}




void
on_clutDrawArea_event_after            (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
	//if (event->type != GDK_BUTTON_PRESS) return;
//fprintf(stderr,"clut event after %0x\n",event->type);
	
}

void
on_drawArea_event_after                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
	char *wname;
	if (event->type != GDK_BUTTON_PRESS) return;
	//fprintf(stderr,"drawArea event_after %0x\n",event->type);
	wname = (char *) gtk_widget_get_name((GtkWidget *) widget);
	set_update_flags(wname);


}


/**********
 *
 *  piccel_gen_clut actions
 *
 *********/

void 
loadArray(double arr[2][6],char *prefix, GtkWidget *widget) {
  int i;
  char sbname[16];

  for (i=0;i<6;i++) {
    sprintf(sbname,"%s%d",prefix,i);
    arr[0][i] = gtk_spin_button_get_value( (GtkSpinButton*) lookup_widget(widget,sbname) );
  }
  for (i=0;i<6;i++) {
    arr[1][i] = arr[0][(i!=5 ? i+1 : 0)] - arr[0][i];
  }
}

void 
setSBarray(double arr[2][6], char *prefix, GtkWidget *widget) {
  int i;
  char sbname[16];

  for (i=0;i<6;i++) {
    sprintf(sbname,"%s%d",prefix,i);
    gtk_spin_button_set_value( (GtkSpinButton*) lookup_widget(widget,sbname), arr[0][i] );
  }
}


double
interpArray(double arr[2][6],double angle) {
  int index;
  double offset;
  angle *= 6.0;
  index = (int) angle;
  offset = angle - d(index);
  return (arr[0][index] + arr[1][index] * offset);
}

void
un_interpArray(double val, double arr[2][6],double angle) {
  int index;
  double offset;
  angle *= 6.0;
  index = (int) angle;
  offset = angle - d(index);
  if (offset < arr[1][index]) {
    arr[0][index] = val;
    arr[1][index] = offset;
  }
}

void
on_piccel_gen_clut_destroy             (GtkObject       *object,
                                        gpointer         user_data)
{
	fprintf(stderr,"piccel_gen_clut destroyed\n");

}

void
drawCurrentDA(GtkWidget *widget, int region) {

  GdkColor gcolor;
  GdkGC *gc;
  int totw,w;
  int h;
  int i;

  double gamma;
  double saturation;
  double luminance;

  GtkWidget *window = gtk_widget_get_toplevel(widget);
  char *parent_title = (char *) gtk_window_get_title(GTK_WINDOW( window )); 
  
  char prof[16];
  char base[16];
  char wname[24];

  region = (region == 5 ? 0 : region +1);

  if (workLUT != (LPLUT) 0) {
    cmsFreeLUT(workLUT);
    workLUT = (LPLUT) 0; // pedantic
  }

  sscanf(parent_title,"%s profile %s CLUT",prof,base);
  sprintf(wname,"%sclut",base);

  workTag = getTag(base[0],(int) (base[3] - '0'),NULL);

  if (workLUT = getLUT( prof[0],workTag )) {
    LPWORD Tbase = workLUT->T;
    LPWORD T;
    int clutpoints = workLUT->cLutPoints;
    int Tunit=3;
    int l_ndx,a_ndx,b_ndx;

    gamma      = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbgamma"));  // / 100.0;
    saturation = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbosat"));
    luminance  = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbvl"));

    gc = gdk_gc_new(widget->window);

    h = widget->allocation.height;
    w = widget->allocation.width;
    totw = w * 6;

    for (i=0; i<w; i++) {
      double color[3];
      double incr = d((w*region)+i)/d(totw);
      gcolor.pixel = 0;
      WORD in[3],out[3];

      color[0] = luminance;
      color[1] = sin(incr*d(2)*M_PI) * saturation;
      color[2] = cos(incr*d(2)*M_PI) * saturation;
      
      lab2xyzA(color);
      xyz2rgbA(color,gamma);
      
      in[0] = (WORD) round (color[1]);   // 1,2,0
      in[1] = (WORD) round (color[2]);
      in[2] = (WORD) round (color[0]);

      cmsEvalLUT(workLUT,in,out);

      //printf("%5.2lf %6d %6d %6d  ->  %6d %6d %6d \n",incr,color[0],color[1],color[2],tcolor[0],tcolor[1],tcolor[2]);

      gcolor.red   = (guint16)  out[0];
      gcolor.green = (guint16)  out[1];
      gcolor.blue  = (guint16)  out[2]; 

      gdk_gc_set_rgb_fg_color (gc,&gcolor);
      gdk_draw_line(widget->window,gc,i,0,i,h);
    }
    
  }
}
void
drawInputDA(GtkWidget *widget, int region) {

  GdkColor gcolor;
  GdkGC *gc;
  int totw,w;
  int h;
  int i;

  double gamma;
  double saturation;
  double luminance;

  //drawCurrentDA(widget,region);
  //return;

  // ***************************

  gamma      = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbgamma"));  // / 100.0;
  saturation = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbosat"));
  luminance  = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbvl"));

  //fprintf(stderr,"gamma %lf  sat  %lf   lum %lf\n",gamma,saturation,luminance);
  gc = gdk_gc_new(widget->window);

  h = widget->allocation.height;
  w = widget->allocation.width;
  totw = w * 6;

  for (i=0; i<w; i++) {
    double color[3];
    double incr = d((w*region)+i)/d(totw);
    gcolor.pixel = 0;

    color[0] = luminance;
    color[1] = sin(incr*d(2)*M_PI) * saturation;
    color[2] = cos(incr*d(2)*M_PI) * saturation;

    lab2xyzA(color);
    xyz2rgbA(color,gamma);

    gcolor.red   = (guint16)  round( color[0] );
    gcolor.green = (guint16)  round( color[1] );
    gcolor.blue  = (guint16)  round( color[2] );

    //fprintf(stderr," a %lf  g %d    b %lf  bl %d\n",color[1],gcolor.green,color[2],gcolor.blue);

    gdk_gc_set_rgb_fg_color (gc,&gcolor);
    gdk_draw_line(widget->window,gc,i,0,i,h);
  }

}

void
updateInputBar(GtkWidget *widget) {
  int i;
  GtkWidget *dwidget;
  for (i=0;i<6;i++) {
    char name[16];
    sprintf(name,"ida%d",i);
    dwidget = lookup_widget(widget,name);
    drawInputDA(dwidget,i);
  }
}

void 
drawTrimNeutralDA(GtkWidget *widget) {

  GdkColor gcolor;
  GdkGC *gc;
  int w,h,i;

  double gamma;
  double trimHA,trimLA,trimHB,trimLB;

  gamma      = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbgamma")); // / 100.0;
  trimHA     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnha"));
  trimLA     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnsa"));
  trimHB     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnhb"));
  trimLB     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnsb"));


  gc = gdk_gc_new(widget->window);

  h = widget->allocation.height;
  w = widget->allocation.width;

  for (i=0; i<w; i++) {
    double color[3];
    gcolor.pixel = 0;

    color[0] = 100.0-(d(i)/d(w))*100.0;
    color[1] = ((trimHA*(w-i-1)) + (trimLA*i))/w;
    color[2] = ((trimHB*(w-i-1)) + (trimLB*i))/w;

    lab2xyzA(color);
    xyz2rgbA(color,gamma);

    gcolor.red   = (guint16)  round( color[0] );
    gcolor.green = (guint16)  round( color[1] );
    gcolor.blue  = (guint16)  round( color[2] );

    //fprintf(stderr," a %lf  g %d    b %lf  bl %d\n",color[1],gcolor.green,color[2],gcolor.blue);

    gdk_gc_set_rgb_fg_color (gc,&gcolor);
    gdk_draw_line(widget->window,gc,i,0,i,h);
  }

}

void
updateNeutralBar(GtkWidget *widget) {
  GtkWidget *dwidget;
  dwidget = lookup_widget(widget,"nda");
  drawTrimNeutralDA(dwidget);
}


void
drawOutputDA(GtkWidget *widget, int region) {

  GdkColor gcolor;
  GdkGC *gc;
  int totw,w;
  int h;
  int i;

  double gamma;
  double saturation;
  double luminance;
  double trimHA,trimLA,trimHB,trimLB;
  double sat,trimA,trimB,angle;


  gamma      = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbgamma")); // / 100.0;
  saturation = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbosat"));
  luminance  = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbvl"));
  trimHA     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnha"));
  trimLA     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnsa"));
  trimHB     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnhb"));
  trimLB     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnsb"));

  //fprintf(stderr,"gamma %lf  sat  %lf   lum %lf\n",gamma,saturation,luminance);

  gc = gdk_gc_new(widget->window);

  h = widget->allocation.height;
  w = widget->allocation.width;
  totw = w * 6;

  for (i=0; i<w; i++) {
    double color[3];
    double incr = d((w*region)+i)/d(totw);
    gcolor.pixel = 0;

    color[0] = luminance;

    sat = saturation * interpArray(satArray,incr) * interpArray(lumArray,incr);  // * interpsat(incr) * interplum(incr)
    angle = incr + interpArray(shiftArray,incr)/10.0; // + interpshift(incr)/10
    trimA = (trimHA * (100.0-luminance) + trimLA*luminance) / 100.0 ;
    color[1] = sin(angle*2.0*M_PI) * sat + trimA;

    trimB = (trimHB * (100.0-luminance) + trimLB*luminance) / 100.0 ;
    color[2] = cos(angle*2.0*M_PI) * sat + trimB;

    lab2xyzA(color);
    xyz2rgbA(color,gamma);

    gcolor.red   = (guint16)  round( color[0] );
    gcolor.green = (guint16)  round( color[1] );
    gcolor.blue  = (guint16)  round( color[2] );

    //fprintf(stderr," a %lf  g %d    b %lf  bl %d\n",color[1],gcolor.green,color[2],gcolor.blue);

    gdk_gc_set_rgb_fg_color (gc,&gcolor);
    gdk_draw_line(widget->window,gc,i,0,i,h);
  }

}


void
updateOutputBar(GtkWidget *widget) {
  int i;
  GtkWidget *dwidget;

  for (i=0;i<6;i++) {
    char name[16];
    sprintf(name,"tda%d",i);
    dwidget = lookup_widget(widget,name);
    drawOutputDA(dwidget,i);
  }
}

gboolean
on_ida0_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  drawInputDA(widget,0);
  return FALSE;
}


gboolean
on_ida1_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  drawInputDA(widget,1);
  return FALSE;
}


gboolean
on_ida2_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  drawInputDA(widget,2);
  return FALSE;
}


gboolean
on_ida3_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  drawInputDA(widget,3);
  return FALSE;
}


gboolean
on_ida4_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  drawInputDA(widget,4);
  return FALSE;
}


gboolean
on_ida5_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  drawInputDA(widget,5);
  return FALSE;
}


void
loadInterpArrays(GtkWidget *widget) {
  loadArray(satArray,"sbsat",widget);
  loadArray(lumArray,"sblum",widget);
  loadArray(shiftArray,"sbcs",widget);
}

void
setInterpArrays(GtkWidget *widget) {
  setSBarray(a_satArray,"sbsat",widget);
  loadArray(satArray,"sbsat",widget);


  setSBarray(a_lumArray,"sblum",widget);
  loadArray(lumArray,"sblum",widget);


  setSBarray(a_shiftArray,"sbcs",widget);
  loadArray(shiftArray,"sbcs",widget);  
}

gboolean
on_tda0_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  loadInterpArrays(widget);
  drawOutputDA(widget,0);
  return FALSE;
}


gboolean
on_tda1_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  loadInterpArrays(widget);
  drawOutputDA(widget,1);
  return FALSE;
}


gboolean
on_tda2_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  loadInterpArrays(widget);
  drawOutputDA(widget,2);
  return FALSE;
}


gboolean
on_tda3_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  loadInterpArrays(widget);
  drawOutputDA(widget,3);
  return FALSE;
}


gboolean
on_tda4_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  loadInterpArrays(widget);
  drawOutputDA(widget,4);
  return FALSE;
}


gboolean
on_tda5_expose_event                   (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  loadInterpArrays(widget);
  drawOutputDA(widget,5);
  return FALSE;
}


void
on_sbsat0_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(satArray,"sbsat",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sblum0_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(lumArray,"sblum",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbcs0_value_changed                 (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(shiftArray,"sbcs",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbsat1_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(satArray,"sbsat",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sblum1_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(lumArray,"sblum",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbcs1_value_changed                 (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(shiftArray,"sbcs",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbsat2_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(satArray,"sbsat",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sblum2_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(lumArray,"sblum",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbcs2_value_changed                 (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(shiftArray,"sbcs",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbsat3_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(satArray,"sbsat",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sblum3_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(lumArray,"sblum",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbcs3_value_changed                 (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(shiftArray,"sbcs",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbsat4_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(satArray,"sbsat",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sblum4_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(lumArray,"sblum",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbcs4_value_changed                 (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(shiftArray,"sbcs",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbsat5_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(satArray,"sbsat",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sblum5_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(lumArray,"sblum",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbcs5_value_changed                 (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  loadArray(shiftArray,"sbcs",(GtkWidget*) spinbutton);
  updateOutputBar((GtkWidget*)spinbutton);
}


void
on_sbosat_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  updateOutputBar((GtkWidget *) spinbutton);
  updateInputBar((GtkWidget *) spinbutton);
}


void
on_sbgamma_value_changed               (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  updateOutputBar((GtkWidget *) spinbutton);
  updateInputBar((GtkWidget *) spinbutton);
  updateNeutralBar((GtkWidget *) spinbutton);
}


void
on_sbtnha_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  updateOutputBar((GtkWidget *) spinbutton);
  updateNeutralBar((GtkWidget *) spinbutton);
}


void
on_sbtnhb_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  updateOutputBar((GtkWidget *) spinbutton);
  updateNeutralBar((GtkWidget *) spinbutton);
}


void
on_sbtnsa_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  updateOutputBar((GtkWidget *) spinbutton);
  updateNeutralBar((GtkWidget *) spinbutton);
}


void
on_sbtnsb_value_changed                (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  updateOutputBar((GtkWidget *) spinbutton);
  updateNeutralBar((GtkWidget *) spinbutton);
}

gboolean
on_nda_expose_event                    (GtkWidget       *widget,
                                        GdkEventExpose  *event,
                                        gpointer         user_data)
{
  drawTrimNeutralDA(widget);
  return FALSE;
}


#define L_LR 0
#define L_AR 1
#define L_BR 2

#define SQR(x) ((x)*(x))

void
on_clutbuildbtn_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{

  // attempted port from labwarp

  int i,j,k;
  int clutpoints,cp1;
  double dcp1;

  double mag,sat2,gam;


	/*
	   see renderClutSlice and follow alg from buildclut.rewrite.1

	   working here.....

	 */
  double gamma;
  double saturation;
  double luminance;
  double trimHA,trimLA,trimHB,trimLB;
  double satv,trimA,trimB,angle;
  double color[3];

  GtkWidget *widget = (GtkWidget *) button;
  GtkWidget *window = gtk_widget_get_toplevel(widget);
  char *parent_title = (char *) gtk_window_get_title(GTK_WINDOW( window )); 
  
  char prof[16];
  char base[16];
  char wname[24];

  if (workLUT != (LPLUT) 0) {
    cmsFreeLUT(workLUT);
    workLUT = (LPLUT) 0; // pedantic
  }

  sscanf(parent_title,"%s profile %s CLUT",prof,base);
  sprintf(wname,"%sclut",base);
  set_update_flags(wname);
    // hack!!!!!!!!!! to get 3 luts updated
  //set_update_flags("b2a1clut");
  //set_update_flags("b2a2clut");
  workTag = getTag(base[0],(int) (base[3] - '0'),NULL);

  gamma      = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbgamma")) ; /// 100.0;
  satv       = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbosat"));
  saturation = satv ; /// 100.0;

  trimHA     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnha"));
  trimLA     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnsa"));
  trimHB     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnhb"));
  trimLB     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnsb"));

  fprintf(stderr,"%s clut rebuild %s gam %lf sat %lf lum %lf trimha %lf trimla %lf trimhb %lf trimlb %lf \n",
	  prof, base,gamma,saturation,luminance,trimHA,trimLA,trimHB,trimLB);
  

  if (workLUT = getLUT( prof[0],workTag )) {
    LPWORD Tbase = workLUT->T;

    clutpoints = workLUT->cLutPoints;
    cp1 = clutpoints-1;
    dcp1 = d(cp1);
    for (k=0; k< clutpoints; k++) {
      for (j=0; j< clutpoints; j++) {
	for (i=0; i< clutpoints; i++) {

	  color[L_LR] = d(k) * (100.0/dcp1);
	  trimA = ((trimHA * color[L_LR]) + (trimLA * 100.0) - color[L_LR]) / 100.0;
	  trimB = ((trimHB * color[L_LR]) + (trimLB * 100.0) - color[L_LR]) / 100.0;
	  color[L_AR] = d(j) * (200.0 / dcp1) + trimA -100.0;
	  color[L_BR] = d(i) * (200.0 / dcp1) + trimB -100.0;
	  
	  if (color[L_BR] == 0.0) {
	    color[L_BR] = 0.00001; // fudge 
	  }
	  angle = atan(color[L_AR]/color[L_BR]);
	  
	  if (color[L_BR] < 0) {
	    angle += M_PI;
	  }

	  if (angle < 0) {
	    angle += (2.0* M_PI);
	  }
	  angle /= (2.0* M_PI);

	  mag = sqrt(SQR(color[L_AR]) + SQR(color[L_BR]));

	  color[L_LR] *= (1.0 - (1.0 - interpArray(lumArray,angle)) * (mag / 100.0));
	  saturation = satv * interpArray(satArray,angle) / 100.0;
	  angle += interpArray(shiftArray,angle) / 10.0;

	  color[L_AR] = mag * sin(angle * 2.0 * M_PI) * saturation;
	  color[L_BR] = mag * cos(angle * 2.0 * M_PI) * saturation;
	  

	  lab2xyzA(color);

	  //fprintf(stderr,"%4d %4d %4d  %5.2f %5.2f %5.2f\n",i,j,k,color[L_LR],color[L_AR],color[L_BR]);

	  xyz2rgbA(color,gamma);

	  //fprintf(stderr,"f %4d %4d %4d  %5.2f %5.2f %5.2f\n",i,j,k,color[L_LR],color[L_AR],color[L_BR]);

	  *Tbase++ = (WORD) round( color[0] );
	  *Tbase++ = (WORD) round( color[1] );
	  *Tbase++ = (WORD) round( color[2] );

	}
      }
    }

  } else {
    fprintf(stderr,"unable to get LUT for %s profile %s\n",prof,base);
  }

  fprintf(stderr,"finished build clut, wl= %lx \n",workLUT);
}


void
on_clutloadset_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *fcdlg;
  
  fcdlg = create_filechooserdialog1 ();
  setdir(fcdlg);
  if (gtk_dialog_run (GTK_DIALOG (fcdlg)) == GTK_RESPONSE_OK) {
    char *filename;
    FILE *infile;
    
    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (fcdlg));
    savedir(filename);
    if ((infile = fopen(filename,"r")) != (FILE*) NULL) {
      char label[32];
      double value;

      while ((fscanf(infile,"%s %lf\n",label,&value)) != EOF) {
	GtkSpinButton *sb;
	if ((sb = (GtkSpinButton*) lookup_widget((GtkWidget *)button,label)) != (GtkSpinButton*) NULL) {
	  gtk_spin_button_set_value(sb,value);
	} else {
	  fprintf(stderr,"could not find spinbutton %s value to set %lf\n",label,value);
	}
      }
    } else {
      sprintf(errstr,"unable to open %s for writing",filename);
      alert();
    }    
  } else {
    printf ("load clut settings: response not OK\n");
  }

  gtk_widget_destroy (fcdlg);

}


void
on_clutsaveset_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *fsdlg;

  char *saveset[] = {
    "sbsat0","sbsat1","sbsat2","sbsat3","sbsat4","sbsat5",
    "sblum0","sblum1","sblum2","sblum3","sblum4","sblum5",
    "sbcs0","sbcs1","sbcs2","sbcs3","sbcs4","sbcs5",
    "sbosat","sbgamma","sbvl","sbtnha","sbtnhb","sbtnsa","sbtnsb",
    ""
  };

  char **ssndx=saveset;
  FILE *outfile;

  fsdlg = create_filesaverdialog();
  setdir(fsdlg);

  if (gtk_dialog_run (GTK_DIALOG (fsdlg)) == GTK_RESPONSE_OK) {
    char *filename;

    filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER(fsdlg));
    savedir(filename);

    if ((outfile = fopen(filename,"w")) != (FILE *) NULL) {
      while (**ssndx != '\0') {
	fprintf(outfile,"%s %lf\n",*ssndx,gtk_spin_button_get_value((GtkSpinButton *) lookup_widget((GtkWidget*)button,*ssndx)));
	ssndx++;
      }
      fclose(outfile);
    } else {
      sprintf(errstr,"unable to open %s for writing",filename);
      alert();
    }
    
  } else {
    printf ("clut save settings: response not OK\n");
  }

  gtk_widget_destroy (fsdlg);

}


void
on_sbvl_value_changed                  (GtkSpinButton   *spinbutton,
                                        gpointer         user_data)
{
  updateOutputBar((GtkWidget *) spinbutton);
  updateInputBar((GtkWidget *) spinbutton);
}

void
on_approximate_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{

  // attempted REVERSE of port from labwarp

  int i,j,k;
  int clutpoints,cp1;
  double dcp1;

  double mag;


	/*

	   working here.....

	 */
  double gamma;
  double sat,lum,ang;

  double trimHA,trimLA,trimHB,trimLB;
  double satv,trimA,trimB,angle;
  double color[3];
  double tcolor[3];

  GtkWidget *widget = (GtkWidget *) button;
  GtkWidget *window = gtk_widget_get_toplevel(widget);
  char *parent_title = (char *) gtk_window_get_title(GTK_WINDOW( window )); 
  
  char prof[16];
  char base[16];
  char wname[24];


  fprintf(stderr,"sorry, approximation de-implemented because color space conversions not sufficiently reversible\n");
  return;

  // ***************************

  if (workLUT != (LPLUT) 0) {
    cmsFreeLUT(workLUT);
    workLUT = (LPLUT) 0; // pedantic
  }

  sscanf(parent_title,"%s profile %s CLUT",prof,base);
  sprintf(wname,"%sclut",base);

  workTag = getTag(base[0],(int) (base[3] - '0'),NULL);

  gamma      = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbgamma")) ; /// 100.0;
  satv       = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbosat"));

  trimHA     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnha"));
  trimLA     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnsa"));
  trimHB     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnhb"));
  trimLB     = gtk_spin_button_get_value((GtkSpinButton *) lookup_widget(widget,"sbtnsb"));

  fprintf(stderr,"%s clut rebuild - approximate vals %s gam %lf sat %lf lum %lf trimha %lf trimla %lf trimhb %lf trimlb %lf \n",
	  prof, base,gamma,satv,lum,trimHA,trimLA,trimHB,trimLB);

  for (i=0;i<6;i++) {
    a_satArray[1][i] = 1.0;
    a_lumArray[1][i] = 1.0;
    a_shiftArray[1][i] = 1.0;
  }

  if (workLUT = getLUT( prof[0],workTag )) {
    LPWORD Tbase = workLUT->T;

    clutpoints = workLUT->cLutPoints;
    cp1 = clutpoints-1;
    dcp1 = d(cp1);
    for (k=0; k< clutpoints; k++) {
      for (j=0; j< clutpoints; j++) {
	for (i=0; i< clutpoints; i++) {

	  color[L_LR] = d(k) * (100.0/dcp1);
	  trimA = ((trimHA * color[L_LR]) + (trimLA * 100.0) - color[L_LR]) / 100.0;
	  trimB = ((trimHB * color[L_LR]) + (trimLB * 100.0) - color[L_LR]) / 100.0;
	  color[L_AR] = d(j) * (200.0 / dcp1) + trimA -100.0;
	  color[L_BR] = d(i) * (200.0 / dcp1) + trimB -100.0;
	  
	  if (color[L_BR] == 0.0) {
	    color[L_BR] = 0.00001; // fudge 
	  }
	  angle = atan(color[L_AR]/color[L_BR]);
	  
	  if (color[L_BR] < 0) {
	    angle += M_PI;
	  }

	  if (angle < 0) {
	    angle += (2.0* M_PI);
	  }
	  angle /= (2.0* M_PI);

	  mag = sqrt(SQR(color[L_AR]) + SQR(color[L_BR]));

	  tcolor[0] = *Tbase++;
	  tcolor[1] = *Tbase++;
	  tcolor[2] = *Tbase++;


	  //fprintf(stderr,"p %4d %4d %4d  %5.2f %5.2f %5.2f\n",i,j,k,tcolor[L_LR],tcolor[L_AR],tcolor[L_BR]);
	  
	  rgb2xyzA(tcolor,gamma);

	  //fprintf(stderr,"%4d %4d %4d  %5.2f %5.2f %5.2f\n",i,j,k,tcolor[L_LR],tcolor[L_AR],tcolor[L_BR]);

	  xyz2labA(tcolor);

	  
	  if (color[L_LR] == 0.0) {
	    color[L_LR] = 0.00001;
	  }
	  if (mag == 0.0) {
	    mag = 0.00001;
	  }
	  lum = tcolor[L_LR] / color[L_LR];
	  lum = 1.0 - lum;
	  lum /= (mag / 100.0);
	  lum = 1.0 - lum;
	  un_interpArray(lum,a_lumArray,angle);


	  ang = atan(tcolor[1] / tcolor[2]) / (2.0*M_PI);
	  sat = tcolor[1] / (mag * sin(ang * 2.0 * M_PI));
	  un_interpArray(sat,a_satArray,angle);
	  
	  ang -= angle;
	  ang *= 10;
	  un_interpArray(ang,a_shiftArray,angle);

	}
      }
    }

    setInterpArrays(widget);
    
  } else {
    fprintf(stderr,"unable to get LUT for %s profile %s for approximation\n",prof,base);
  }

  fprintf(stderr,"finished approximating clut, wl= %lx \n",workLUT);

}

